"""
Class for Directed Acyclic Graphs (DAGs).
"""
import itertools
import logging

from .admg import ADMG
from .cg import CG
from .ug import UG

logger = logging.getLogger(__name__)


class UndefinedDAGOperation(Exception):
    pass


class DAG(ADMG, CG):
    def __init__(self, vertices=[], di_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        """

        super().__init__(vertices=vertices, di_edges=di_edges, **kwargs)
        logger.debug("DAG")

    def d_separated(self, X, Y, separating_set=list()):
        """
        Computes d-separation for set `X` and set `Y` given `separating_set`

        :param X: first vertex set
        :param Y: second vertex set
        :param separating_set: separating set, default list()
        :return: boolean result of d-separation
        """
        if type(X) is str:
            X = [X]
        if type(Y) is str:
            Y = [Y]
        if type(separating_set) is str:
            separating_set = [separating_set]
        for x, y in itertools.product(X, Y):
            if not self._d_separated(x, y, separating_set):
                return False
        return True

    def _d_separated(self, X, Y, separating_set):
        """
        Determine whether two vertices are d-separated given other vertices. Also handles conditional DAGs

        :param X: first vertex
        :param Y: second vertex
        :param separating_set: list of given vertices

        :return: boolean result of d-separation
        """
        ancestral_vars = [X, Y] + list(separating_set)

        # create a new subgraph of the ancestors of X, Y, and separating vertices
        ancestral_subgraph = self.subgraph(self.ancestors(ancestral_vars))
        ancestral_vertices = list(ancestral_subgraph.vertices)
        ancestral_edges = list(ancestral_subgraph.di_edges)

        # add fixed variables which are not in X or Y to sepset
        for V in ancestral_subgraph.vertices.values():
            if V.fixed and V.name not in X and V.name not in Y:
                separating_set.append(V.name)

        # if both vertices are fixed, result is undefined
        if (
            ancestral_subgraph.vertices[X].fixed
            and ancestral_subgraph.vertices[Y].fixed
        ):
            raise UndefinedDAGOperation(
                "{0} and {1} are fixed, so d-separation is undefined.".format(
                    X, Y
                )
            )

        for Vi, Vj in itertools.combinations(ancestral_vertices, 2):
            # fixed vertices are connected
            if (
                ancestral_subgraph.vertices[Vi].fixed
                and ancestral_subgraph.vertices[Vj].fixed
            ):
                ancestral_edges.append((Vi, Vj))

            ancestral_subgraph.vertices[Vi].fixed = False
            ancestral_subgraph.vertices[Vj].fixed = False

        # retrieves all combinations of the graph's vertices
        pairs_of_vertices = [
            list(pair) for pair in itertools.combinations(ancestral_vertices, 2)
        ]

        # checks for common children between any pairs of vertices
        # if a pair of vertices has common children, an undirected edge connects the vertices
        for Vi, Vj in pairs_of_vertices:
            children_i = set(ancestral_subgraph.children(Vi))
            children_j = set(ancestral_subgraph.children(Vj))
            common_children = children_i.intersection(children_j)
            if len(common_children) > 0:
                ancestral_edges.append((Vi, Vj))

        # removes given vertices from the graph
        for vertex in separating_set:
            ancestral_vertices.remove(vertex)

        # removes any edges from the graph that include any of the given vertices
        for edge in ancestral_edges[:]:
            if edge[0] in separating_set or edge[1] in separating_set:
                ancestral_edges.remove(edge)

        # creates a new undirected graph from the updated vertices and edges
        augmented_graph = UG(ancestral_vertices, ancestral_edges)

        # checks if vertex Y is in the block of vertex X
        Y_block = augmented_graph.block(X)

        return Y not in Y_block



#!/usr/bin/env python3
"""
Implements the Nested Markov parameterization for acyclic directed mixed graphs over binary variables.

The code relies on the following papers:

[ER12] Evans, R. J., & Richardson, T. S. (2012). Maximum likelihood fitting of acyclic directed mixed graphs to binary data. arXiv preprint arXiv:1203.3479.
[ER19] Evans, R. J., & Richardson, T. S. (2019). Smooth, identifiable supermodels of discrete DAG models with latent variables. Bernoulli, 25(2), 848-876.

"""
import logging
from collections import ChainMap, OrderedDict, namedtuple

import numpy as np
import pandas as pd
from scipy import optimize

from ..identification import one_line
from ..utils import powerset

logger = logging.getLogger(__name__)


def permutations(n, k=2):
    """
    Computes tuples of all permutations of n variables with cardinality k.

    :param n: number of variables
    :param k: cardinality of each variable
    :return:
    """
    return [
        tuple(x)
        for x in np.rollaxis(
            np.mgrid[(slice(0, k, None),) * n], 0, n + 1
        ).reshape((k**n, n))
    ]


def recursive_partition_heads(B, intrinsic_dict):
    """
    Partition an arbitrary (sub)set of vertices B (in V) into recursive heads

    :param B: arbitrary possibly empty subset of vertices
    :param intrinsic_dict: map of intrinsic sets to heads and tails of that set

    """
    if len(B) == 0:  # if B is null set
        heads = []
    else:
        rH = get_recursive_heads(intrinsic_dict)
        B_star = B
        heads = []

        while len(B_star) > 0:  # some empty set condition

            pB = [set(element) for element in list(set(powerset(B_star, 0)))]
            rH_intersection_pB = [element for element in pB if element in rH]

            tau = maximal_heads(rH_intersection_pB, intrinsic_dict)
            heads.append(tau)

            tau_list = [
                item for sublist in [list(i) for i in tau] for item in sublist
            ]
            B_star = B_star.difference(set(tau_list))

        heads = [
            item for sublist in [list(i) for i in heads] for item in sublist
        ]
        heads = {
            frozenset(i) for i in heads
        }  # convert to frozen set so it can be used as a key

    return heads


def maximal_heads(list_H, intrinsic_dict):
    """
    Returns maximal heads for a set of heads. Only defined if all heads in list_H (list) are in IHT, ie, they are all heads of intrinsic sets
    """
    # create a dictionary that is the reverse mapping of heads:intrinsic sets
    reverse_dict = {
        frozenset(head_tail[0]): iset
        for iset, head_tail in intrinsic_dict.items()
    }
    list_iset = [
        reverse_dict[frozenset(x)] for x in list_H
    ]  # the only sets we care about

    # order both lists by number of elements in each intrinsic set (descending)
    list_iset_nelem = [len(i) for i in list_iset]
    ord_iset = [
        x for _, x in sorted(zip(list_iset_nelem, list_iset), reverse=True)
    ]
    ord_H = [x for _, x in sorted(zip(list_iset_nelem, list_H), reverse=True)]

    max_heads = []

    for ix, iset in enumerate(ord_iset):
        bigger_lst = ord_iset[
            :ix
        ]  # all intrinsic sets "larger" than this one; does not include self.
        if len(bigger_lst) == 0:
            max_heads.append(ord_H[ix])
        else:
            temp = [
                iset.issubset(big_set) for big_set in bigger_lst
            ]  # can make this better. If iset is a subset of one bigger set we dont care about the others
            if not any(temp):
                max_heads.append(ord_H[ix])

    return max_heads


def get_recursive_heads(intrinsic_dict):
    """
    Compute all heads of intrinsic sets

    :param intrinsic_dict: mapping of intrinsic sets of some graph to heads and tails
    :return: list of all heads
    """

    return [v[0] for k, v in intrinsic_dict.items()]


def initialize_q_vector(intrinsic_dict):
    """
    Generates the q_vector, a dictionary mapping (heads: frozenset, tails: tuple, value of tail: tuple) to parameter. Default q parameter values are 1/2^#(head), for each head.

    :param intrinsic_dict: mapping of intrinsic sets of some graph to heads and tails
    :return q_vector:
    """
    q_vector = OrderedDict()
    ordered_values = sorted(
        [tuple(tuple(sorted(y)) for y in x) for x in intrinsic_dict.values()]
    )
    for head, tail in ordered_values:
        n_head = len(head)
        n_tail = len(tail)
        if n_tail == 0:
            q_vector[(frozenset(head), tuple(tail), ())] = 1 / (2**n_head)
        else:

            for permutation in permutations(n_tail):
                q_vector[
                    (frozenset(head), tuple(tail), tuple(permutation))
                ] = 1 / (2**n_head)

    return q_vector


def get_heads_tails_map(intrinsic_dict):
    """
    Get mapping of heads to tails from a mapping of intrinsic sets

    :param intrinsic_dict: mapping of intrinsic sets of some graph to heads and tails
    :return:
    """
    return {
        frozenset(value[0]): tuple(sorted(value[1]))
        for key, value in intrinsic_dict.items()
    }


def compute_likelihood_district(
    nu_dict, q_vector, district, heads_tails_map, intrinsic_dict
):
    """
    Compute likelihood directly using Equation (1) of [ER12].

    This likelihood is not recommended for use in maximization as it is more efficiently expressed using M and P computations.

    :param nu_dict: a dictionary representing a single observation of variables to value
    :param q_vector: dictionary mapping the head, tail, tail value to parameter value
    :param district: district of graph
    :param heads_tails_map: mapping of heads to tails
    :param intrinsic_dict: mapping of intrinsic set to heads and tails
    :return:
    """
    """
    """
    # find all subsets of that district -  some of these will be B's (based on what nu is)
    pset = [
        frozenset(i) for i in list(powerset(district, 0))
    ]  # convert to frozen set so you can set a dictionary with these keys later
    ## maybe for each district send in the powerset as a dictionary?

    # compute partition of heads for each of these subsets
    partition_head_dict = {
        B: recursive_partition_heads(B, intrinsic_dict) for B in pset
    }

    O_j = [
        key
        for key, value in nu_dict.items()
        if (value == 0 and key in district)
    ]
    B_j = [
        set(pset_item)
        for pset_item in pset
        if all(element in list(pset_item) for element in O_j)
    ]  # subset of D_j but superset of O_j

    lld = 0
    for setB in B_j:
        partition = partition_head_dict[frozenset(setB)]
        term = 1
        for head in partition:
            # if setB is empty, partition is also empty and term = 1 (defined)
            if len(heads_tails_map[head]) == 0:
                term = term * q_vector[(head, (), ())]
            else:
                tail = heads_tails_map[head]
                nu_tail = tuple(
                    [nu_dict[tail_element] for tail_element in tail]
                )
                term = term * q_vector[head, tail, nu_tail]

        lld = lld + (-1) ** (len(setB) - len(O_j)) * term

    return lld


def compute_M(G, partition_head_dict, district, heads_tails_map, terms):
    """
    Compute M matrix as given in Section 3.1 of [ER12]

    :param G: ADMG graph
    :param partition_head_dict: dictionary mapping every subset of given district to constituent heads
    :param district: a district of the graph
    :param heads_tails_map: dictionary mapping heads to tails
    :param terms: list of terms of given district
    :return:
    """
    district = sorted(district)
    pset = [
        tuple(sorted(i)) for i in list(powerset(district, 0))
    ]  # convert to frozen set so you can set a dictionary with these keys later
    all_nu = [
        OrderedDict(zip(sorted(list(G.vertices)), tuple(permutation)))
        for permutation in permutations(len(G.vertices))
    ]
    all_term_dicts = []

    for nu_dict in all_nu:

        O_j = [
            key
            for key, value in nu_dict.items()
            if (value == 0 and key in district)
        ]
        B_j = [
            (pset_item)
            for pset_item in pset
            if all(element in list(pset_item) for element in O_j)
        ]  # subset of D_j but superset of O_j
        terms_dict = OrderedDict(zip(terms, [0] * len(terms)))
        for setB in B_j:
            all_tails = list()
            for item in partition_head_dict[setB]:
                tail = heads_tails_map[item]
                if len(tail) > 0:
                    all_tails.extend(tail)
            all_tails = tuple(sorted(list(set(all_tails))))
            values = tuple(nu_dict[x] for x in all_tails)

            terms_dict[(frozenset(setB), all_tails, values)] = (-1) ** (
                len(setB) - len(O_j)
            )

        assert list(terms_dict.keys()) == terms, terms

        all_term_dicts.append(list(terms_dict.values()))

    return np.array(all_term_dicts)


def compute_P(partition_head_dict, q_vector_keys, heads_tails_map, terms):
    """
    Computes P matrix as given in Section 3.1 of [ER12]

    :param partition_head_dict: dictionary mapping every subset of given district to constituent heads
    :param q_vector_keys: list of parameter names
    :param heads_tails_map: dictionary mapping heads to tails
    :param terms: list of all terms for a given district
    :return:
    """
    all_P = []
    for term in terms:
        P_dict = OrderedDict(zip(q_vector_keys, [0] * len(q_vector_keys)))

        head = tuple(sorted(term[0]))
        all_tails = term[1]
        all_values = term[2]
        for item in partition_head_dict[head]:
            ix = [
                i
                for i in range(len(all_tails))
                if all_tails[i] in heads_tails_map[item]
            ]
            tail = tuple([all_tails[i] for i in ix])
            value = tuple([all_values[i] for i in ix])
            P_dict[(frozenset(item), tail, value)] = 1
        all_P.append(list(P_dict.values()))

    return np.array(all_P)


def compute_all_M_and_P(G, intrinsic_dict):
    """
    Computes for each district of a graph, the corresponding M and P matrices required to compute the expression

    .. math::
        p(q) = \prod_j M_j \exp(P_j \log q_j)

    See Section 3.1 of [ER12] for further details.

    :param G: ADMG
    :param intrinsic_dict: mapping of intrinsic set to heads and tails
    :return:
    """
    heads_tails_map = get_heads_tails_map(intrinsic_dict=intrinsic_dict)
    q_vector_keys = list(initialize_q_vector(intrinsic_dict).keys())
    Ms = []
    Ps = []
    districts = [frozenset(x) for x in G.districts]

    for D_j in districts:
        terms = compute_terms(D_j, intrinsic_dict, heads_tails_map)
        partition_head_dict = compute_partition_head_dict(
            intrinsic_dict=intrinsic_dict, district=D_j
        )
        M = compute_M(
            G=G,
            partition_head_dict=partition_head_dict,
            district=D_j,
            heads_tails_map=heads_tails_map,
            terms=terms,
        )
        # Select q parameters whose heads are in district D_j
        selected_q_vector_keys = [
            k for k in q_vector_keys if set(k[0]).issubset((D_j))
        ]
        P = compute_P(
            partition_head_dict=partition_head_dict,
            q_vector_keys=selected_q_vector_keys,
            heads_tails_map=heads_tails_map,
            terms=terms,
        )
        Ms.append(M)
        Ps.append(P)
    return Ms, Ps


def compute_district_bool_map(q_vector_keys, districts):
    """

    :param q_vector_keys:
    :param districts:
    :return:
    """
    selected_keys = []
    districts = [frozenset(x) for x in districts]
    for D_j in districts:
        selected_keys.append(
            [True if set(k[0]).issubset(D_j) else False for k in q_vector_keys]
        )
    return OrderedDict(zip(districts, selected_keys))


def compute_partition_head_dict(intrinsic_dict, district):
    """
    Compute partition head dictionary. Maps every subset of a district to its constituent maximal recursive heads

    :param intrinsic_dict: dictionary mapping intrinsic sets to (heads, tails) of that set
    :param district: district of graph
    :return:
    """
    B_j = sorted(
        [tuple(sorted(i)) for i in list(powerset(district, 0))]
    )  # convert to frozen set so you can set a dictionary with these keys later
    partition_head_dict = {
        B: recursive_partition_heads(frozenset(B), intrinsic_dict) for B in B_j
    }

    return partition_head_dict


def compute_terms(district, intrinsic_dict, heads_tails_map):
    """
    Computes list of terms (product of q parameters) and the partition head dictionary.
    Each term is a tuple of (frozenset(all heads), tuple(all tails), tuple(values of all tails)).

    :param district: district of graph
    :param intrinsic_dict: dictionary mapping intrinsic sets to (heads, tails) of that set
    :param heads_tails_map: dictionary mapping heads to tails
    :return:
    """
    # find all subsets of that district -  some of these will be B's (based on what nu is)
    B_j = sorted(
        [tuple(sorted(i)) for i in list(powerset(district, 0))]
    )  # convert to frozen set so you can set a dictionary with these keys later

    # compute partition of heads for each of these subsets
    partition_head_dict = compute_partition_head_dict(
        intrinsic_dict=intrinsic_dict, district=district
    )
    heads_tails_values = list()
    for B in B_j:
        all_tails = list()
        for item in partition_head_dict[B]:
            tail = heads_tails_map[item]
            if len(tail) > 0:
                all_tails.extend(tail)
        all_tails = sorted(list(set(all_tails)))
        n_tail = len(all_tails)
        tail_permutations = permutations(n=n_tail)
        for value in tail_permutations:
            heads_tails_values.append(
                (frozenset(B), tuple(all_tails), tuple(value))
            )

    return sorted(heads_tails_values)


def _compute_prob_single_district(q, M, P):
    """
    Computes the probability of a single district as per Eqn (3) of [ER12]

    :param q: q parameters involved in district
    :param M: M matrix for district
    :param P: P matrix for district
    :return:
    """
    prob = M @ np.exp(P @ np.log(q))
    return prob


def compute_q_indices_by_district(q_vector_keys, districts):
    """
    Computes a boolean indexing array that indicates which q parameters are involved in which districts.

    :param q_vector_keys:
    :param districts:
    :return:
    """
    indices = [
        [True if k[0].issubset(district) else False for k in q_vector_keys]
        for district in districts
    ]
    return indices


def _compute_prob(q, q_indices, districts, Ms, Ps):
    """
    Compute full probability using matrix operations

    :param q: numpy array of q parameters (obtained from q_vector)
    :param q_indices: numpy array of booleans to select subset of active q for each district
    :param districts: list of districts
    :param Ms: list of M matrices
    :param Ps: list of P matrices
    :return:
    """
    prob = 1
    for i, (district, M, P) in enumerate(zip(districts, Ms, Ps)):
        new_q = q[q_indices[i]]
        prob *= _compute_prob_single_district(
            q=new_q, M=M, P=P
        )  # M @ np.exp(P @ np.log(new_q))
    return prob


def _compute_likelihood(q, counts, q_indices, districts, Ms, Ps):
    """
    Computes full log-likelihood

    :param q: numpy array of q parameters (obtained from q_vector)
    :param counts: n(V=v) for each value assignment v of variables V, obtained in sorted order
    :param q_indices: numpy array of booleans to select subset of active q for each district
    :param districts: list of districts
    :param Ms: list of M matrices
    :param Ps: list of P matrices
    :return:
    """
    prob = _compute_prob(q, q_indices, districts, Ms, Ps)
    lld = np.sum(counts * np.log(prob))
    return lld


def _compute_partial_neg_log_likelihood(x, counts, A, b):
    """
    Returns negative partial log-likelihood for optimization.
    This function is implicitly defined w.r.t. a variable v, such that the parameter x consists of all
    parameters in q whose heads contain v. This selected variable defines the corresponding A, b matrices.
    The probability vector is then computed as prob = A @ x - b

    :param x: subset of q
    :param counts: n(V=v) for each value assignment v of variables V, obtained in sorted order
    :param A: A matrix as per Eqn (4) of [ER12]
    :param b: b vector as per Eqn (4) of [ER12]
    """

    logger.debug("prob")
    logger.debug(A @ x - b)
    logger.debug("logprob")
    logger.debug(np.log(A @ x - b))
    lld = np.sum(counts * np.log(A @ x - b))
    return -lld


def _compute_jac_neg_log_partial_likelihood(x, counts, A, b):
    """
    Returns negative jacobian of partial log-likelihood for optimization

    :param x: subset of q
    :param counts: n(V=v) for each value assignment v of variables V, obtained in sorted order
    :param A: A matrix as per Eqn (4) of [ER12]
    :param b: b vector as per Eqn (4) of [ER12]
    :return:
    """
    jac = np.sum(A * (counts * 1 / (A @ x - b)).reshape(-1, 1), axis=0)
    return -jac


def _compute_hess_neg_log_partial_likelihood(x, counts, A, b):
    """
    Compute hessian of partial log-likelihood. Because the partial likelihood is a linear function, the hessian is zero.

    :param x: subset of q
    :param counts: n(V=v) for each value assignment v of variables V, obtained in sorted order
    :param A: A matrix as per Eqn (4) of [ER12]
    :param b: b vector as per Eqn (4) of [ER12]
    :return:
    """
    return np.zeros((x.shape[0], x.shape[0]))


def compute_theta_reindexed_bool_map(q_vector_keys, districts):
    """
    Creates a mapping from a variable to a boolean indexing vector.

    This boolean vector indexes only parameters whose heads are involved in the district of that variable.
    It selects parameters which have heads containing that variable.

    Used to construct A and b matrices for partial likelihood.

    :param q_vector_keys: list of q_vector keys
    :param districts: list of districts
    :return:
    """
    theta_reindexed_bool_map = dict()
    for district in districts:
        selected_q_vector_keys = [
            k for k in q_vector_keys if set(k[0]).issubset((district))
        ]
        for v in district:
            theta_reindexed_bool_map[v] = np.array(
                [
                    True if v in item[0] else False
                    for i, item in enumerate(selected_q_vector_keys)
                ]
            )
    return theta_reindexed_bool_map


def compute_theta_bool_map(q_vector_keys, variables):
    """
    Compute map from variable to boolean indexing vector of q_parameters which have heads containing that variable.
    In this map, the indices are not reindexed by district. The boolean vector selects parameters which have heads containing that variable.

    :param q_vector_keys: list of q_vector keys
    :param variables: list of variables
    :return:
    """
    theta_bool_map = {
        v: np.array(
            [
                True if v in item[0] else False
                for i, item in enumerate(q_vector_keys)
            ]
        )
        for v in variables
    }
    return theta_bool_map


def construct_A_b(variable, q, theta_reindexed_bool_map, M, P):
    """
    Constructs A and b matrices (eqn 4, Evans and Richardson 2013) for constraining parameters of given variable,
    in district which admits matrices M, P

    :param variable: name of variable
    :param q_vector: q_vector in OrderedDict format
    :param M: M matrix
    :param P: P matrix
    :return:
    """
    theta_bool_ix = theta_reindexed_bool_map[variable]
    theta_ix = np.argwhere(theta_bool_ix).ravel()

    comp_bool_ix = ~theta_bool_ix
    comp_ix = np.argwhere(comp_bool_ix).ravel()

    comp_val = q[comp_ix]
    P_theta = P[:, theta_ix]
    P_comp = P[:, comp_ix]

    b = -M[:, np.sum(P_theta, axis=1) == 0] @ np.exp(
        P_comp[np.sum(P_theta, axis=1) == 0, :] @ np.log(comp_val)
    )

    A = np.concatenate(
        [
            M[:, P[:, theta_ix[i]] == 1]
            @ np.exp(
                P[np.ix_(P[:, theta_ix[i]] == 1, comp_bool_ix)]
                @ np.log(comp_val)
            ).reshape(-1, 1)
            for i in range(len(theta_ix))
        ],
        axis=1,
    )

    return A, b


def process_data(df, count_variable=None):
    """
    :param data: pandas DataFrame columns of variables and rows of observations
    :param count_variable: optional name of counting variable, if data is provided as a summary table
    :return: a vector of counts, ordered in ascending order of v for p(V=v)
    """
    if count_variable is None:
        colnames = sorted(list(df.columns))
    else:
        colnames = sorted(list(set(df.columns) - {count_variable}))

    Counts = namedtuple("Counts", colnames)

    count_dict = OrderedDict()
    for permutation in permutations(len(colnames)):
        count_dict[(Counts(**dict(zip(colnames, permutation))))] = 0

    for tup in df.itertuples(index=False):
        if count_variable is None:
            permutation_tuple = Counts(**dict(zip(tup._fields, list(tup))))
            count_dict[permutation_tuple] += 1
        else:
            fields = list(tup._fields)
            bad_index = fields.index(count_variable)
            fields.remove(count_variable)
            values = list(tup)
            del values[bad_index]
            permutation_tuple = Counts(**dict(zip(fields, values)))
            count_dict[permutation_tuple] += getattr(tup, count_variable)

    counts = np.array(list(count_dict.values()))

    return counts


def compute_counterfactual_binary_parameters(G, q_vector, x_dict, y_dict):
    """
    Computes a counterfactual (interventional quantity) p(Y=y | do(X=x)) using nested Markov parameters and graph

    :param G: ADMG
    :param q_vector: A dictionary of nested Markov parameters
    :param x_dict: A dictionary of treatment variables to treatment values
    :param y_dict: A dictionary of outcome variables to outcome values
    :returns: Computed probability p(Y=y | do(X=x))
    """

    Ga = G.fix(list(x_dict))
    Gy_star = Ga.subgraph(Ga.ancestors(list(y_dict)))
    y_star = set(Gy_star.vertices)
    y_star_not_y = y_star - set(y_dict) - set(x_dict)
    y_star_not_y_values = [
        dict(zip(list(y_star_not_y), x))
        for x in permutations(len(y_star_not_y), 2)
    ]
    intrinsic_dict, _ = G.get_intrinsic_sets_and_heads()

    heads_tails_map = get_heads_tails_map(intrinsic_dict=intrinsic_dict)
    prob_tot = 0
    for value_dict in y_star_not_y_values:
        prob = 1
        all_vals_dict = dict(ChainMap(x_dict, y_dict, value_dict))
        for district in Gy_star.districts:

            prob_inc = compute_likelihood_district(
                all_vals_dict,
                q_vector,
                district,
                heads_tails_map,
                intrinsic_dict,
            )
            prob *= prob_inc
        prob_tot += prob

    return prob_tot


class BinaryNestedModel:
    """
    Estimates parameters of nested binary model using the iterative maximum likelihood algorithm (Algorithm 1) of [ER12].
    Performs scipy.minimize constrained minimization of negative log-likelihood.
    """

    def __init__(self, graph):
        self.graph = graph
        self.fitted_params = None

    def fit(self, X, q_vector=None, tol=1e-8, *args, **kwargs):
        """
        Fits the binary nested model. Let N the number of observations, M the number of variables.

        :param X: Either a N x M pandas DataFrame where each row represents an observation, or an 2 ** M x (M+1) pandas DataFrame, where each row represents the observation through a count variable
        :param args:
        :param kwargs:
        :return:
        """
        if type(X) == pd.core.frame.DataFrame:
            if sorted(list(X.columns)) == sorted(self.graph.vertices):
                counts = process_data(X)
            elif len(list(X.columns)) == len(self.graph.vertices) + 1:
                count_variable = list(
                    set(X.columns) - set(self.graph.vertices)
                )[0]
                counts = process_data(X, count_variable=count_variable)
        else:
            counts = X

        return self._fit(counts, q_vector, tol, *args, **kwargs)

    def _fit(self, counts, q_vector, tol, *args, **kwargs):
        intrinsic_dict, _ = self.graph.get_intrinsic_sets_and_heads()
        districts = [frozenset(x) for x in self.graph.districts]
        variables = list(self.graph.vertices)
        if q_vector is None:
            q_vector = initialize_q_vector(intrinsic_dict)
        Ms, Ps = compute_all_M_and_P(
            G=self.graph, intrinsic_dict=intrinsic_dict
        )
        variable_district_map = dict(
            ChainMap(
                *[{v: district for v in district} for district in districts]
            )
        )
        Ms_map = dict(zip(districts, Ms))
        Ps_map = dict(zip(districts, Ps))
        district_bool_map = compute_district_bool_map(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )
        theta_bool_map = compute_theta_bool_map(
            q_vector_keys=list(q_vector.keys()), variables=variables
        )
        theta_reindexed_bool_map = compute_theta_reindexed_bool_map(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )
        q_indices = compute_q_indices_by_district(
            q_vector_keys=list(q_vector.keys()), districts=districts
        )

        q = np.array(list(q_vector.values()))
        logger.debug("starting q: {}".format(str(q)))
        diff = np.inf

        while diff > tol:
            lld_old = _compute_likelihood(
                q=q,
                counts=counts,
                q_indices=q_indices,
                districts=districts,
                Ms=Ms,
                Ps=Ps,
            )
            logger.debug("current likelihood: {}".format(lld_old))
            for v in variables:
                logger.debug("current variable: {}".format(v))
                district = variable_district_map[v]
                q_district = q[district_bool_map[district]]
                M = Ms_map[district]
                P = Ps_map[district]
                A, b = construct_A_b(
                    variable=v,
                    q=q_district,
                    theta_reindexed_bool_map=theta_reindexed_bool_map,
                    M=M,
                    P=P,
                )
                theta = q[theta_bool_map[v]]
                constraint = optimize.LinearConstraint(
                    A=A, lb=b, ub=np.inf, keep_feasible=True
                )
                res = optimize.minimize(
                    _compute_partial_neg_log_likelihood,
                    x0=theta,
                    args=(counts, A, b),
                    constraints=[constraint],
                    method="trust-constr",
                    jac=_compute_jac_neg_log_partial_likelihood,
                    hess=_compute_hess_neg_log_partial_likelihood,
                )
                q[theta_bool_map[v]] = res.x
                logger.debug("q after {}: {}".format(v, str(q)))

            lld_new = _compute_likelihood(
                q=q,
                counts=counts,
                q_indices=q_indices,
                districts=districts,
                Ms=Ms,
                Ps=Ps,
            )
            diff = lld_new - lld_old
            logger.debug("q: {}".format(str(q)))
            logger.debug("likelihood increase: {}".format(str(diff)))

        self.fitted_params = OrderedDict(zip(q_vector.keys(), q))
        self.lld = lld_new

        return self

    def estimate(
        self, treatment_dict=None, outcome_dict=None, check_identified=True
    ):
        """
        Estimates p(Y(a)=y) for treatment A=a, outcome Y=y using the binary nested Markov parameters.

        :param treatment_dict: dict of treatment variables to values
        :param outcome_dict: dict of outcome variables to values
        :param check_identified: boolean to check that effect is identified, default True
        :return: interventional probability p(Y(a)) if identified, else return None
        """
        if outcome_dict is None:
            raise ValueError("Outcome dict is not specified!")
        if treatment_dict is None:
            treatment_dict = dict()
        assert self.fitted_params is not None, "Fit model first!"
        if check_identified:
            assert one_line.OneLineID(
                self.graph,
                treatments=list(treatment_dict),
                outcomes=list(outcome_dict),
            ).id(), "Causal effect is not identified!"

        return compute_counterfactual_binary_parameters(
            G=self.graph.copy(),
            q_vector=self.fitted_params,
            x_dict=treatment_dict,
            y_dict=outcome_dict,
        )



"""
Class for automated derivation of influence functions.
"""

import copy

from .counterfactual_mean import CausalEffect


class AutomatedIF:
    """
    IF for a single treatment and single outcome.
    """

    def __init__(self, graph, treatment, outcome):
        """
        Constructor.

        :param graph: ADMG corresponding to substantive knowledge/model.
        :param treatment: name of vertex corresponding to the intervention.
        :param outcome: name of vertex corresponding to the outcome.
        """

        # without loss of generality, focus on ancestors of Y
        self.graph = copy.deepcopy(graph)
        self.treatment = treatment
        self.outcome = outcome

        self.ace = CausalEffect(self.graph, self.treatment, self.outcome)

        if self.ace.strategy == "Not ID":
            raise RuntimeError("Query is not identified.")

        self.beta_primal_ = ""
        self.beta_dual_ = ""
        self.nonparametric_if_ = ""
        self.eff_if_ = ""

        if self.ace.strategy == "a-fixable":
            self._format_augmented_ipw()

        elif self.ace.strategy == "p-fixable":
            self._format_augmented_primal_ipw()

        else:
            raise NotImplementedError(
                "Influence functions are available only when the treatment is "
                + "a-fixable or p-fixable"
            )

    def _format_density(self, V, cond_set, intervened=False, expectation=False):
        """
        Provide appropriate formatting of a p(V | conditioning set).

        :param V: vertex of interest.
        :param cond_set: set of vertices to condition on.
        :param expectation: whether it is a conditional expectation instead of density i.e. E[ | ].
        :return: string formatted appropriately.
        """

        density = ""

        if len(cond_set) == 0:
            if expectation:
                density = "E[{}]".format(V)
            else:
                density = "p({})".format(V)
        else:
            if expectation:
                density = "E[{}|{}]".format(V, ",".join(cond_set))
            else:
                density = "p({}|{})".format(V, ",".join(cond_set))

        if intervened:
            density = density.replace(
                self.treatment, self.treatment + "=" + self.treatment.lower()
            )

        return density

    def _format_augmented_ipw(self):
        """
        Format the IF for augmented IPW and efficient augmented IPW if possible.

        :return: None.
        """

        # get the Markov pillow of T
        mp_T = self.graph.markov_pillow([self.treatment], self.ace.p_order)

        # format the propensity score/IPW
        ipw_form = "I({}={}) x 1/".format(
            self.treatment, self.treatment.lower()
        )
        ipw_form += self._format_density(self.treatment, mp_T)
        self.beta_primal_ += ipw_form + " x {}".format(self.outcome)

        # format the outcome regression
        self.beta_dual_ += self._format_density(
            self.outcome,
            mp_T.union({self.treatment}),
            expectation=True,
            intervened=True,
        )

        # nonparametric IF
        self.nonparametric_if_ = (
            ipw_form
            + " x ({} - {})".format(self.outcome, self.beta_dual_)
            + " + "
            + self.beta_dual_
            + " - Ψ"
        )

        # eff IF if nonparametric saturated is just nonparametric IF
        if self.graph.nonparametric_saturated():
            self.eff_if_ = self.nonparametric_if_

        # otherwise do projections if graph is mb-shielded
        elif self.ace.is_mb_shielded:
            contributions = []
            for V in set(self.graph.vertices).difference({self.treatment}):
                mpV = self.graph.markov_pillow([V], self.ace.p_order)
                VmpV = set([V]).union(mpV)
                contributions.append(
                    self._format_density("βprimal", VmpV, expectation=True)
                    + " - "
                    + self._format_density("βprimal", mpV, expectation=True)
                )
            self.eff_if_ = " + ".join(contributions)

        # otherwise we do not know how to compute the efficient IF
        else:
            self.eff_if_ = "Cannot compute, graph is not mb-shielded."

    def _format_augmented_primal_ipw(self):
        """
        Format the IF for augmented primal IPW and efficient augmented primal IPW if possible.

        :return:
        """

        # C := pre-treatment vars and L := post-treatment vars in district of treatment
        C = self.graph.pre([self.treatment], self.ace.p_order)
        post = set(self.graph.vertices).difference(C)
        L = post.intersection(self.graph.district(self.treatment))

        self.beta_primal_ = "I({}={}) x 1/[".format(
            self.treatment, self.treatment.lower()
        )
        primal_terms = []
        for Li in L.difference([self.outcome]):
            mpLi = self.graph.markov_pillow([Li], self.ace.p_order)
            primal_terms.append(self._format_density(Li, mpLi))
        self.beta_primal_ += "".join(primal_terms) + "] x Σ_{} ".format(
            self.treatment
        )
        self.beta_primal_ += "".join(primal_terms)

        if self.outcome in L:
            mpY = self.graph.markov_pillow([self.outcome], self.ace.p_order)
            self.beta_primal_ += " x " + self._format_density(
                self.outcome, mpY, expectation=True
            )
        else:
            self.beta_primal_ += " x {}".format(self.outcome)

        # M := inverse Markov pillow of the treatment
        M = set(
            [
                m
                for m in self.graph.vertices
                if self.treatment
                in self.graph.markov_pillow([m], self.ace.p_order)
            ]
        )
        M = M.difference(self.graph.district(self.treatment))
        dual_terms = []
        for Mi in M.difference([self.outcome]):
            mpMi = self.graph.markov_pillow([Mi], self.ace.p_order)
            dual_fixed_term = self._format_density(Mi, mpMi, intervened=True)
            dual_random_term = self._format_density(Mi, mpMi)
            dual_terms.append(
                "[" + dual_fixed_term + "/" + dual_random_term + "]"
            )
        self.beta_dual_ = " x ".join(dual_terms)

        if self.outcome in M:
            mpY = self.graph.markov_pillow([self.outcome], self.ace.p_order)
            self.beta_dual_ += " x " + self._format_density(
                self.outcome, mpY, expectation=True, intervened=True
            )
        else:
            self.beta_dual_ += " x {}".format(self.outcome)

        # nonparametric IF
        contributions = []
        for V in self.graph.vertices:

            preV = self.graph.pre([V], self.ace.p_order)
            VpreV = set([V]).union(preV)

            if V in M:
                contributions.append(
                    self._format_density("βprimal", VpreV, expectation=True)
                    + " - "
                    + self._format_density("βprimal", preV, expectation=True)
                )
            elif V in L:
                contributions.append(
                    self._format_density("βdual", VpreV, expectation=True)
                    + " - "
                    + self._format_density("βdual", preV, expectation=True)
                )
            else:
                contributions.append(
                    self._format_density(
                        "βprimal or βdual", VpreV, expectation=True
                    )
                    + " - "
                    + self._format_density(
                        "βprimal or βdual", preV, expectation=True
                    )
                )
        self.nonparametric_if_ = " + ".join(contributions)

        # eff IF if nonparametric saturated is just nonparametric IF
        if self.graph.nonparametric_saturated():
            self.eff_if_ = self.nonparametric_if_

        elif self.ace.is_mb_shielded:
            contributions = []
            for V in self.graph.vertices:

                mpV = self.graph.markov_pillow([V], self.ace.p_order)
                VmpV = set([V]).union(mpV)

                if V in M:
                    contributions.append(
                        self._format_density("βprimal", VmpV, expectation=True)
                        + " - "
                        + self._format_density("βprimal", mpV, expectation=True)
                    )
                elif V in L:
                    contributions.append(
                        self._format_density("βdual", VmpV, expectation=True)
                        + " - "
                        + self._format_density("βdual", mpV, expectation=True)
                    )
                else:
                    contributions.append(
                        self._format_density(
                            "βprimal or βdual", VmpV, expectation=True
                        )
                        + " - "
                        + self._format_density(
                            "βprimal or βdual", mpV, expectation=True
                        )
                    )
            self.eff_if_ = " + ".join(contributions)

        # otherwise we do not know how to compute the efficient IF
        else:
            self.eff_if_ = "Cannot compute, graph is not mb-shielded."



"""
Base class for all graphs.

TODO: Add error checking
"""

import collections
import copy

from .vertex import Vertex


class Graph:
    def __init__(
        self,
        vertices=[],
        di_edges=set(),
        bi_edges=set(),
        ud_edges=set(),
        **kwargs,
    ):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        :param bi_edges: iterable of tuples of bidirected edges i.e. (X, Y) = X <-> Y.
        :param ud_edges: iterable of tuples of undirected edges i.e. (X, Y) = X - Y.
        """
        assert not kwargs, "Unrecognised kwargs: {}".format(kwargs)

        # initialize vertices
        if isinstance(vertices, collections.abc.Mapping):
            if all(
                isinstance(x, str) or isinstance(x, int) or x is None
                for x in vertices.values()
            ):
                self.vertices = {
                    v: Vertex(name=v, cardinality=c)
                    for v, c in vertices.items()
                }
            elif all(isinstance(x, Vertex) for x in vertices.values()):
                self.vertices = vertices
            else:
                raise ValueError("Unrecognized mapping of vertices")
        else:
            if all(isinstance(x, Vertex) for x in vertices):
                self.vertices = {x.name: x for x in vertices}
            elif all(isinstance(x, str) for x in vertices):
                self.vertices = {v: Vertex(v) for v in vertices}
            else:
                raise ValueError("Unrecognized iterable of vertices")

        # initialize edges
        self.di_edges = set()
        self.bi_edges = set()
        self.ud_edges = set()

        # read in directed edges
        # explicit reference to Graph so that overridden functions
        # aren't called that make calls to recompute districts etc.
        for parent, child in di_edges:
            Graph.add_diedge(self, parent, child)

        # read in bidirected edges
        for sib1, sib2 in bi_edges:
            Graph.add_biedge(self, sib1, sib2)

        # read in undirected edges
        for neb1, neb2 in ud_edges:
            Graph.add_udedge(self, neb1, neb2)

    def add_vertex(self, name, cardinality=None, **kwargs):
        """
        Add a vertex to the graph.

        :param name: name of vertex.
        :return: None.
        """

        if name not in self.vertices:
            self.vertices[name] = Vertex(
                name=name, cardinality=cardinality, **kwargs
            )

    def delete_vertex(self, name):
        """
        Delete a vertex from the graph. Additionally removes any edges from other vertices into the
        given vertex

        :param name: name of vertex.
        :return: None.
        """
        vertex = self.vertices[name]
        for child in [x.name for x in vertex.children]:
            self.delete_diedge(name, child)
        for parent in [x.name for x in vertex.parents]:
            self.delete_diedge(parent, name)
        for sib in [x.name for x in vertex.siblings]:
            self.delete_biedge(sib, name)
        for neb in [x.name for x in vertex.neighbors]:
            self.delete_udedge(neb, name)

        del self.vertices[name]

    def add_diedge(self, parent, child):
        """
        Add a directed edge to the graph.

        :param parent: tail of edge.
        :param child: head of edge.
        :return: None.
        """
        if (parent, child) not in self.di_edges:
            self.di_edges.add((parent, child))
            self.vertices[parent].children.add(self.vertices[child])
            self.vertices[child].parents.add(self.vertices[parent])

    def delete_diedge(self, parent, child):
        """
        Deleted given directed edge from the graph.

        :param parent: tail of edge.
        :param child: head of edge.
        :return: None.
        """

        self.di_edges.remove((parent, child))
        self.vertices[parent].children.remove(self.vertices[child])
        self.vertices[child].parents.remove(self.vertices[parent])

    def add_biedge(self, sib1, sib2):
        """
        Add a bidirected edge to the graph.

        :param sib1: endpoint 1 of edge.
        :param sib2: endpoint 2 of edge.
        :return: None.
        """
        if (sib1, sib2) and (sib2, sib1) not in self.bi_edges:

            self.bi_edges.add((sib1, sib2))
            self.vertices[sib1].siblings.add(self.vertices[sib2])
            self.vertices[sib2].siblings.add(self.vertices[sib1])

    def delete_biedge(self, sib1, sib2):
        """
        Delete given bidirected edge from the graph.

        :param sib1: endpoint 1 of edge.
        :param sib2: endpoint 2 of edge.
        :return: None.
        """

        try:
            self.bi_edges.remove((sib1, sib2))
        except KeyError:
            self.bi_edges.remove((sib2, sib1))

        self.vertices[sib1].siblings.remove(self.vertices[sib2])
        self.vertices[sib2].siblings.remove(self.vertices[sib1])

    def has_biedge(self, sib1, sib2):
        """
        Check existence of a bidirected edge.

        :param sib1: endpoint 1 of edge.
        :param sib2: endpoint 2 of edge.
        :return: boolean result of existence.
        """

        if (sib1, sib2) in self.bi_edges or (sib2, sib1) in self.bi_edges:
            return True
        return False

    def has_udedge(self, neb1, neb2):
        """
        Check existence of an undirected edge.

        :param neb1: endpoint 1 of edge.
        :param neb2: endpoint 2 of edge.
        :return: boolean result of existence.
        """

        if (neb1, neb2) in self.ud_edges or (neb2, neb1) in self.ud_edges:
            return True
        return False

    def add_udedge(self, neb1, neb2):
        """
        Add an undirected edge to the graph.

        :param neb1: endpoint 1 of edge.
        :param neb2: endpoint 2 of edge.
        :return: None.
        """
        if (neb1, neb2) and (neb2, neb1) not in self.ud_edges:

            self.ud_edges.add((neb1, neb2))
            self.vertices[neb1].neighbors.add(self.vertices[neb2])
            self.vertices[neb2].neighbors.add(self.vertices[neb1])

    def delete_udedge(self, neb1, neb2):
        """
        Delete given undirected edge from the graph.

        :param neb1: endpoint 1 of edge.
        :param neb2: endpoint 2 of edge.
        :return: None.
        """

        try:
            self.ud_edges.remove((neb1, neb2))
        except KeyError:
            self.ud_edges.remove((neb2, neb1))

        self.vertices[neb1].neighbors.remove(self.vertices[neb2])
        self.vertices[neb2].neighbors.remove(self.vertices[neb1])

    #### GENEALOGICAL HELPERS ####
    def parents(self, vertices):
        """
        Get parents of a vertex or set of vertices.

        :param vertices: iterable of vertex names.
        :return: set of parents.
        """

        parents = set()
        if type(vertices) == str:
            vertices = [vertices]
        for v in vertices:
            for p in self.vertices[v].parents:
                parents.add(p.name)
        return parents

    def children(self, vertices):
        """
        Get children of a vertex or set of vertices.

        :param vertices: iterable of vertex names.
        :return: set of children.
        """

        children = set()
        if type(vertices) == str:
            vertices = [vertices]
        for v in vertices:
            for c in self.vertices[v].children:
                children.add(c.name)
        return children

    def neighbors(self, vertices):
        """
        Get neighbors of a vertex or set of vertices.

        :param vertices: iterable of vertex names.
        :return: set of neighbors.
        """

        neighbors = set()
        if type(vertices) == str:
            vertices = [vertices]
        for v in vertices:
            for n in self.vertices[v].neighbors:
                neighbors.add(n.name)
        return neighbors

    def siblings(self, vertices):
        """
        Get siblings of a vertex or set of vertices.

        :param vertices: vertex name or iterable of vertex names.
        :return: set of neighbors.
        """

        siblings = set()
        if type(vertices) == str:
            vertices = [vertices]
        for v in vertices:
            for s in self.vertices[v].siblings:
                siblings.add(s.name)
        return siblings

    def ancestors(self, vertices):
        """
        Get the ancestors of a vertex or set of vertices.

        :param vertices: single vertex name or iterable of vertex names to find ancestors for.
        :return: set of ancestors.
        """

        if type(vertices) == str:
            vertices = [vertices]
        ancestors = set()

        visit_stack = list([self.vertices[v] for v in vertices])

        while visit_stack:
            v = visit_stack.pop()
            ancestors.add(v.name)
            visit_stack.extend(p for p in v.parents if p not in ancestors)

        return ancestors

    def descendants(self, vertices):
        """
        Get the descendants of a vertex or set of vertices.

        :param vertices: single vertex name or iterable of vertex names to find descendants for.
        :return: set of descendants.
        """

        if type(vertices) == str:
            vertices = [vertices]
        descendants = set()

        visit_stack = list([self.vertices[v] for v in vertices])

        while visit_stack:
            v = visit_stack.pop()
            descendants.add(v.name)
            visit_stack.extend(c for c in v.children if c not in descendants)

        return descendants

    def _bfs_directed_paths(self, source, sink):
        """
        Use BFS to find directed paths from a source vertex to sink vertices.

        :param source: name of a single source.
        :param sink: iterable of possibly multiple sinks.
        :return: list of directed paths.
        """

        sink_vertices = [self.vertices[v] for v in sink]
        queue = [(self.vertices[source], [])]

        while queue:

            (current_vertex, edge_path) = queue.pop(0)
            for c in current_vertex.children:
                if c in sink_vertices:
                    yield edge_path + [(current_vertex.name, c.name)]
                else:
                    queue.append(
                        (c, edge_path + [(current_vertex.name, c.name)])
                    )

    def directed_paths(self, source, sink):
        """
        Get all directed paths between sets of source vertices and sink vertices.

        :param source: a set of vertices that serve as the source.
        :param sink: a set of vertices that serve as the sink.
        :return: list of directed paths.
        """

        # we use BFS for finding all paths
        directed_paths = []
        for u in source:
            directed_paths += self._bfs_directed_paths(u, sink)
        return directed_paths

    def copy(self):
        """
        Returns copy of a graph
        """
        return copy.deepcopy(self)

    def topological_sort(self):
        """
        Perform a topological sort from roots (parentless nodes)
        to leaves, as ordered by directed edges on the graph.

        :return: list corresponding to a valid topological order.
        """

        # create a copy of the graph
        G = copy.deepcopy(self)

        # initialize roots -- vertices that have no incoming edges
        roots = [v for v in self.vertices if len(G.parents([v])) == 0]
        top_order = []

        # iterate until we have explored all nodes
        while roots:

            # pick a current root
            r = roots.pop()
            top_order.append(r)

            # iterate over all its children
            children = G.children([r])
            for c in children:

                # delete the edge
                G.delete_diedge(r, c)

                # if r was the only remaining parent, this is now a root
                if len(G.parents([c])) == 0:
                    roots.append(c)

        # return the order
        return top_order

    def pre(self, vertices, top_order):
        """
        Find all nodes prior to the given set of vertices under a topological order.

        :param vertices: iterable of vertex names.
        :param top_order: a valid topological order.
        :return: list corresponding to the order up until the given vertices.
        """

        # find all elements that are previous in the topological order
        # by iterating over the order until we encounter one of the vertices
        pre = []
        for v in top_order:
            if v in vertices:
                break
            pre.append(v)

        return pre

    # def post(self, vertices, top_order):
    #     """
    #     Find all nodes that succeed the given set of vertices under a topological order.
    #
    #     :param vertices: iterable of vertex names.
    #     :param top_order: a valid topological order.
    #     :return: list corresponding to the order up until the given vertices.
    #     """
    #
    #     # find all elements that are previous in the topological order
    #     # by iterating over the order until we encounter one of the vertices
    #     post = []
    #     for v in top_order:
    #         if v in vertices:
    #             break
    #         pre.append(v)
    #
    #     return post
    def subgraph(self, vertices):
        """
        Return a subgraph on the given vertices (i.e. a graph containing only
        the specified vertices and edges between them).

        :param vertices: set containing names of vertices in the subgraph.
        :return: a new Graph object corresponding to the subgraph.
        """
        subgraph = copy.deepcopy(self)
        for v in set(self.vertices) - set(vertices):
            subgraph.delete_vertex(v)
        return subgraph

    def draw(self, direction=None):
        """
        Visualize the graph.

        :return : dot language representation of the graph.
        """
        from graphviz import Digraph

        dot = Digraph()

        # set direction from left to right if that's preferred
        if direction == "LR":
            dot.graph_attr["rankdir"] = direction

        for v in self.vertices.values():
            dot.node(
                v.name,
                shape="square" if v.fixed else "plaintext",
                height=".5",
                width=".5",
            )

        for parent, child in self.di_edges:
            # special clause for SWIGs
            if parent.lower() == child:
                dot.edge(parent, child, style="invis")
            else:
                dot.edge(parent, child, color="blue")
        for sib1, sib2 in self.bi_edges:
            dot.edge(sib1, sib2, dir="both", color="red")
        for neb1, neb2 in self.ud_edges:
            dot.edge(neb1, neb2, dir="none", color="brown")

        return dot



import functools
import logging

import numpy as np
from pgmpy.factors.discrete import TabularCPD
from pgmpy.inference.EliminationOrder import (
    MinFill,
    MinNeighbors,
    MinWeight,
    WeightedMinFill,
)
from pgmpy.models import BayesianNetwork

logger = logging.getLogger(__name__)


def variable_elimination(
    bayesian_network, variables, elimination_order_alg="minfill"
):
    """
    Implements the exact inference variable elimination algorithm to compute marginals of a Bayesian
    Network.

    :param bayesian_network: An ananke.BayesianNetwork representing the graph
    :param variables: A list of variables representing the desired marginal
    :param elimination_order_alg: Name of elimination order algorithm (supplied by
    pgmpy.inference.EliminationOrder)

    """
    elimination_order_algs = {
        "minfill": MinFill,
        "weightedminfill": WeightedMinFill,
        "minweight": MinWeight,
        "minneighbors": MinNeighbors,
    }
    elimination_order = elimination_order_algs.get(elimination_order_alg)
    if elimination_order is None:
        raise ValueError(
            f"Invalid choice of ordering: {elimination_order_alg}. Select one from {list(elimination_order_algs)}"
        )

    # Convert bayesian_network into pgmpy.BayesianNetwork, compute elimination ordering
    pgmpy_network = BayesianNetwork()
    pgmpy_network.add_nodes_from(list(bayesian_network.vertices))
    pgmpy_network.add_edges_from(bayesian_network.di_edges)
    random_cpds = list()

    random_cpds = [
        TabularCPD(
            vertex,
            cpd.variable_card,
            np.random.rand(*cpd.get_values().shape),
            cpd.variables[1:],
            cpd.cardinality[1:],
        )
        for vertex, cpd in bayesian_network.cpds.items()
    ]
    pgmpy_network.add_cpds(*random_cpds)

    ordering = elimination_order(pgmpy_network).get_elimination_order(
        set(bayesian_network.vertices) - set(variables)
    )
    # Create factors from each cpd in bayesian network
    factors = [v.to_factor() for v in bayesian_network.cpds.values()]

    for var in ordering:
        logger.debug(f"eliminating {var}")
        selected_factors = list()
        other_factors = list()
        for factor in factors:

            if var in factor.variables:
                selected_factors.append(factor)
            else:
                other_factors.append(factor)

        tau_factor = functools.reduce(
            lambda x, y: x.product(y, inplace=False), selected_factors
        ).marginalize([var], inplace=False)

        factors = other_factors + [tau_factor]
    final_factor = functools.reduce(
        lambda x, y: x.product(y, inplace=False), factors
    )

    return final_factor



import unittest

from ananke.graphs import UG


class TestUG(unittest.TestCase):
    def test_ug_can_be_initialised(self):
        ud_edges = [("A", "B")]
        vertices = ["A", "B"]
        ug = UG(vertices, ud_edges=ud_edges)
        self.assertTrue(UG)
        # test empty BG
        UG()


if __name__ == "__main__":
    unittest.main()



"""
Class for acyclic directed mixed graphs (ADMGs) and conditional ADMGs (CADMGs).
"""
import copy
import itertools
import logging
import warnings
from typing import Union

from ananke.utils import powerset

from .ig import IG
from .sg import SG
from .ug import UG

logger = logging.getLogger(__name__)


class UndefinedADMGOperation(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def latent_project_single_vertex(vertex, graph):
    """
    Latent project one vertex from graph

    :param vertex: Name of vertex to be projected
    :param graph: ADMG
    :returns:
    """

    di_edges = graph.di_edges
    bi_edges = graph.bi_edges

    retained_vertices = set(graph.vertices) - {vertex}
    retained_di_edges = [
        x
        for x in di_edges
        if x[0] in retained_vertices and x[1] in retained_vertices
    ]

    retained_bi_edges = [
        x
        for x in bi_edges
        if x[0] in retained_vertices and x[1] in retained_vertices
    ]

    # Construct all directed edge projections
    new_di_edges = [
        (p.name, c.name)
        for p in graph.vertices[vertex].parents
        for c in graph.vertices[vertex].children
    ]

    # Construct all bidirected edge projections

    new_bi_edges = [
        tuple(sorted([s.name, c.name]))
        for s in graph.vertices[vertex].siblings
        for c in graph.vertices[vertex].children
    ] + [
        tuple(sorted([a.name, b.name]))
        for a, b in itertools.combinations(graph.vertices[vertex].children, r=2)
    ]
    final_bi_edges = retained_bi_edges + new_bi_edges

    G = ADMG(
        vertices=set(retained_vertices),
        di_edges=set(retained_di_edges + new_di_edges),
        bi_edges=set(final_bi_edges),
    )

    return G


class ADMG(SG):
    """
    Class for creating and manipulating (conditional) acyclic directed mixed graphs (ADMGs/CADMGs).
    """

    def __init__(self, vertices=[], di_edges=set(), bi_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        :param bi_edges: iterable of tuples of bidirected edges i.e. (X, Y) = X <-> Y.
        """

        # initialize vertices in ADMG
        super().__init__(
            vertices=vertices, di_edges=di_edges, bi_edges=bi_edges, **kwargs
        )

    def markov_pillow(self, vertices, top_order):
        """
        Get the Markov pillow of a set of vertices. That is,
        the Markov blanket of the vertices given a valid topological order
        on the graph.

        :param vertices: iterable of vertex names.
        :param top_order: a valid topological order.
        :return: set corresponding to Markov pillow.
        """

        # get the subgraph corresponding to the vertices and nodes prior to them
        pre = self.pre(vertices, top_order)
        Gsub = self.subgraph(pre + list(vertices))

        # Markov pillow is the Markov blanket (dis(v) union pa(dis(v)) setminus v)
        # in this subgraph
        pillow = set()
        for v in vertices:
            pillow = pillow.union(Gsub.district(v))
        pillow = pillow.union(Gsub.parents(pillow))
        return pillow - set(vertices)

    def markov_blanket(self, vertices):
        """
        Get the Markov blanket of a set of vertices.

        :param vertices: iterable of vertex names.
        :return: set corresponding to Markov blanket.
        """

        blanket = set()
        for v in vertices:
            blanket = blanket.union(self.district(v))
        blanket = blanket.union(self.parents(blanket))
        return blanket - set(vertices)

    @property
    def fixed(self):
        """
        Returns all fixed nodes in the graph.

        :return:
        """
        fixed_vertices = []
        for v in self.vertices:
            if self.vertices[v].fixed:
                fixed_vertices.append(v)

        return fixed_vertices

    def is_subgraph(self, other):
        """
        Check that this graph is a subgraph of other, meaning it has  a subset of edges and nodes of the other.

        :param other: an object of the ADMG class.
        :return: boolean indicating whether the statement is True or not.
        """
        if (
            set(self.vertices).issubset(set(other.vertices))
            and set(self.di_edges).issubset(set(other.di_edges))
            and set(self.bi_edges).issubset(set(other.bi_edges))
        ):
            return True
        return False

    def is_ancestral_subgraph(self, other):
        """
        Check that this graph is an ancestral subgraph of the other.
        An ancestral subgraph over variables S and intervention b G(S(b)) of a larger graph G(V(b)) is defined as a
        subgraph, such that ancestors of each node s in S with respect to the graph G(V(b_i)) are contained in S.

        :param other: an object of the ADMG class.
        :return: boolean indicating whether the statement is True or not.
        """
        if not self.is_subgraph(other):
            return False

        for v in self.vertices:
            self_parents = set([item.name for item in self.vertices[v].parents])
            other_parents = set(
                [item.name for item in other.vertices[v].parents]
            )
            if self_parents != other_parents:
                return False

        return True

    def reachable_closure(self, vertices):
        """
        Obtain reachable closure for a set of vertices.

        :param vertices: set of vertices to get reachable closure for.
        :return: set corresponding to the reachable closure, the fixing order for vertices
                 outside of the closure, and the CADMG corresponding to the closure.
        """

        # initialize set of vertices that must still be fixed
        remaining_vertices = (
            set(self.vertices)
            - set(vertices)
            - set(v for v in self.vertices if self.vertices[v].fixed)
        )
        fixing_order = []  # keep track of the valid fixing order
        fixed = True  # flag to track that a vertex was successfully fixed in a given pass
        G = copy.deepcopy(self)

        # keep iterating over remaining vertices until there are no more or we failed to fix
        while remaining_vertices and fixed:

            fixed = False

            # check if any remaining vertex van be fixed
            for v in remaining_vertices:

                # fixability check
                if len(G.descendants([v]).intersection(G.district(v))) == 1:
                    G.fix([v])
                    remaining_vertices.remove(v)
                    fixing_order.append(v)
                    fixed = True  # flag that we succeeded
                    break  # stop the current pass over vertices

        # compute final reachable closure based on vertices successfully fixed
        reachable_closure = set(G.vertices) - set(
            v for v in G.vertices if G.vertices[v].fixed
        )

        # return the reachable closure, the valid order, and the resulting CADMG
        return reachable_closure, fixing_order, G

    def fixable(self, vertices):
        """
        Check if there exists a valid fixing order and return such
        an order in the form of a list, else returns an empty list.

        :param vertices: set of vertices to check fixability for.
        :return: a boolean indicating whether the set was fixable and a valid fixing order as a stack.
        """

        # keep track of vertices still left to fix (ignoring fixed vertices)
        # and initialize a fixing order
        G = copy.deepcopy(self)
        remaining_vertices = set(vertices) - set(self.fixed)
        fixing_order = []
        fixed = True  # flag to check if we fixed a variable on each pass

        # while we have more vertices to fix, and were able to perform a fix
        while remaining_vertices and fixed:

            fixed = False

            for v in remaining_vertices:

                # Check if any nodes are reachable via -> AND <->
                # by looking at intersection of district and descendants
                if len(G.descendants([v]).intersection(G.district(v))) == 1:
                    G.fix([v])
                    remaining_vertices.remove(v)
                    fixing_order.append(v)
                    fixed = True
                    break

            # if unsuccessful, return failure and
            # fixing order up until point of failure
            if not fixed:
                return False, fixing_order

        # if fixing vertices was successful, return success
        # and the fixing order
        return True, fixing_order

    def subgraph(self, vertices):
        """
        Computes subgraph given a set of vertices.

        Recomputes districts, since these may change when vertices are removed.

        :param vertices: iterable of vertices
        :return: subgraph
        """
        subgraph = super().subgraph(vertices)
        subgraph._calculate_districts()

        return subgraph

    def get_intrinsic_sets(self):
        """
        Computes intrinsic sets (and returns the fixing order for each intrinsic set).

        :returns: list of intrinsic sets and fixing orders used to reach each one
        """

        # create an intrinsic set graph and obtain the intrinsic sets + valid fixing orders leading to them
        ig = IG(copy.deepcopy(self))
        intrinsic_sets = ig.get_intrinsic_sets()
        fixing_orders = ig.iset_fixing_order_map

        return intrinsic_sets, fixing_orders

    def get_intrinsic_sets_and_heads(self):
        """
        Computes intrinsic sets mapped to a tuple of heads and tails of that intrinsic set, and fixing orders for each one.

        :returns: tuple of dict of intrinsic sets to heads and tails, and fixing orders for each intrinsic set

        """
        ig = IG(copy.deepcopy(self))
        intrinsic_sets = ig.get_intrinsic_sets()
        fixing_orders = ig.iset_fixing_order_map
        heads = []
        tails = []
        for intrinsic_set in intrinsic_sets:
            G_sub = self.subgraph(intrinsic_set)
            head = frozenset(
                {
                    s
                    for s in intrinsic_set
                    if len(G_sub.vertices[s].children) == 0
                }
            )
            tail = frozenset(self.parents(intrinsic_set))
            heads.append(head)
            tails.append(tail)
        return dict(zip(intrinsic_sets, zip(heads, tails))), fixing_orders

    def maximal_arid_projection(self):
        """
        Get the maximal arid projection that encodes the same conditional independences and
        Vermas as the original ADMG. This operation is described in Acyclic
        Linear SEMs obey the Nested Markov property by Shpitser et al 2018.

        :return: An ADMG corresponding to the maximal arid projection.
        """

        vertices, di_edges, bi_edges = self.vertices, [], []

        # keep a cached dictionary of reachable closures and ancestors
        # for efficiency purposes
        reachable_closures = {}
        ancestors = {v: self.ancestors([v]) for v in vertices}

        # iterate through all vertex pairs
        for a, b in itertools.combinations(vertices, 2):

            # decide which reachable closure needs to be computed
            # and compute it if one vertex is an ancestor of another
            u, v, rc = None, None, None
            if a in ancestors[b]:
                u, v = a, b
            elif b in ancestors[a]:
                u, v = b, a

            # check parent condition and add directed edge if u is a parent of the reachable closure
            added_edge = False
            if u:
                if v not in reachable_closures:
                    reachable_closures[v] = self.reachable_closure([v])[0]
                rc = reachable_closures[v]
                if u in self.parents(rc):
                    di_edges.append((u, v))
                    added_edge = True

            # if neither are ancestors of each other we need to compute
            # the reachable closure of set {a, b} and check if it is
            # bidirected connected
            if not added_edge:
                rc, _, cadmg = self.reachable_closure([a, b])
                for district in cadmg.districts:
                    if rc <= district:
                        bi_edges.append((a, b))

        return ADMG(vertices=vertices, di_edges=di_edges, bi_edges=bi_edges)

    def mb_shielded(self):
        """
        Check if the ADMG is a Markov blanket shielded ADMG. That is, check if
        two vertices are non-adjacent only when they are absent from each others'
        Markov blankets.

        :return: boolean indicating if it is mb-shielded or not.
        """

        # iterate over all pairs of vertices
        for Vi, Vj in itertools.combinations(self.vertices, 2):
            # check if the pair is not adjacent
            if not (
                Vi in self.siblings([Vj])
                or (Vi, Vj) in self.di_edges
                or (Vj, Vi) in self.di_edges
            ):
                # if one is in the Markov blanket of the other, then it is not mb-shielded
                if Vi in self.markov_blanket([Vj]) or Vj in self.markov_blanket(
                    [Vi]
                ):
                    return False
        return True

    def nonparametric_saturated(self):
        """
        Check if the nested Markov model implied by the ADMG is nonparametric saturated.
        The following is an implementation of Algorithm 1 in Semiparametric Inference for
        Causal Effects in Graphical Models with Hidden Variables (Bhattacharya, Nabi & Shpitser 2020)
        which was shown to be sound and complete for this task.

        :return: boolean indicating if it is nonparametric saturated or not.
        """

        # iterate over all pairs of vertices
        for Vi, Vj in itertools.combinations(self.vertices, 2):

            # check if there is no dense inducing path between Vi and Vj
            # and return not NPS if either of the checks fail
            if not (
                Vi in self.parents(self.reachable_closure([Vj])[0])
                or Vj in self.parents(self.reachable_closure([Vi])[0])
                or Vi in self.reachable_closure([Vi, Vj])[2].district(Vj)
            ):
                return False
        return True

    def m_separated(self, X, Y, separating_set=list()):
        """
        Computes m-separation for set `X` and set `Y` given `separating_set`

        :param X: first vertex set
        :param Y: second vertex set
        :param separating_set: separating set
        :return: boolean result of m-separation
        """
        if type(X) is str:
            X = [X]
        if type(Y) is str:
            Y = [Y]
        for x, y in itertools.product(X, Y):
            if not self._m_separated(x, y, separating_set):
                return False
        return True

    def _m_separated(self, X, Y, separating_set):
        """
        Determine whether `X` and `Y` vertices are m-separated given `separating_set`. Also works on ADMGs which have fixed vertices.

        :param X: first vertex
        :param Y: second vertex
        :param separating_set: list of given vertices

        :return: boolean result of m-separation
        """

        ancestral_vars = [X, Y] + list(separating_set)

        # create a new subgraph of the ancestors of vertex1, vertex2, and given vertices
        ancestral_subgraph = self.subgraph(self.ancestors(ancestral_vars))

        augmented_graph_vertices = list(ancestral_subgraph.vertices)
        augmented_graph_di_edges = list(ancestral_subgraph.di_edges)

        for V in ancestral_subgraph.vertices.values():
            if V.fixed and V.name != X and V.name != Y:
                separating_set.append(V.name)

        # if both vertices are fixed, result is undefined
        if (
            ancestral_subgraph.vertices[X].fixed
            and ancestral_subgraph.vertices[Y].fixed
        ):
            raise UndefinedADMGOperation(
                "{0} and {1} are fixed, so m-separation is undefined.".format(
                    X, Y
                )
            )

        for Vi, Vj in itertools.combinations(augmented_graph_vertices, 2):
            # fixed vertices are connected
            if (
                ancestral_subgraph.vertices[Vi].fixed
                and ancestral_subgraph.vertices[Vj].fixed
            ):
                augmented_graph_di_edges.append((Vi, Vj))

            ancestral_subgraph.vertices[Vi].fixed = False
            ancestral_subgraph.vertices[Vj].fixed = False

        #
        ancestral_subgraph._calculate_districts()

        for Vi, Vj in itertools.combinations(augmented_graph_vertices, 2):

            # checks for common children between any pairs of vertices
            children_i = set(ancestral_subgraph.children([Vi]))

            markov_blanket_Vi = ancestral_subgraph.markov_blanket([Vi])

            # connects vertices if one is in the markov blanket of the other vertex, or markov blanket of child of other vertex
            if (
                Vj in markov_blanket_Vi
                or Vj in ancestral_subgraph.markov_blanket(children_i)
            ):
                augmented_graph_di_edges.append((Vi, Vj))

        # removes given vertices from the graph
        for vertex in augmented_graph_vertices[:]:
            if vertex in separating_set:
                augmented_graph_vertices.remove(vertex)

        # removes any edges from the graph that include any of the given vertices
        for edge in augmented_graph_di_edges[:]:
            if edge[0] in separating_set or edge[1] in separating_set:
                augmented_graph_di_edges.remove(edge)

        # creates a new undirected graph from the updated vertices and edges
        augmented_graph = UG(augmented_graph_vertices, augmented_graph_di_edges)

        # checks if vertex 2 is in the block of vertex 1
        Y_block = augmented_graph.block(X)
        return Y not in Y_block

    def latent_projection(self, retained_vertices):
        """
        Computes latent projection.

        :param retained_vertices: list of vertices to retain after latent projection
        :returns: Latent projection containing retained vertices
        """

        vertices = self.vertices
        di_edges = self.di_edges
        bi_edges = self.bi_edges

        projected_vertices = list(set(vertices) - set(retained_vertices))
        G = self.copy()

        for vertex in projected_vertices:
            G = latent_project_single_vertex(vertex=vertex, graph=G)

        return G

    def canonical_dag(self, cardinality: Union[int, str] = None):
        """
        Computes the canonical DAG from the ADMG by inserting a variable in place of each bidirected edge.

        For each bidirected edge 'X <-> Y', the inserted variable will take names of the form
        'U_XY', the original bidirected edge is removed, and new directed edges 'U_X_Y -> Y' and
        'U_X_Y -> X' are inserted. The variable names are in lexicographic order.

        :params cardinality: The cardinality of the inserted variables
        :type cardinality: Union[int, str]
        :returns: A directed acyclic graph
        :rtype: DAG


        """
        from .dag import DAG

        if cardinality is None:
            warnings.warn(
                "Warning: cardinality of latent variables not set. Please set the cardinality if intending to use ananke.models.discrete functionality."
            )

        G = DAG(vertices=list(self.vertices), di_edges=self.di_edges)
        for _x, _y in self.bi_edges:
            x, y = tuple(sorted([_x, _y]))
            new_var = "U_{}_{}".format(x, y)
            G.add_vertex(name=new_var, cardinality=cardinality)
            G.add_diedge(new_var, x)
            G.add_diedge(new_var, y)

        for v in self.vertices:
            if self.vertices[v].fixed:
                G.vertices[v].fixed = True

        return G



import unittest

from ananke.graphs import Graph


class TestGraph(unittest.TestCase):
    def test_graph_can_be_initialised(self):
        ud_edges = [("A", "B")]
        vertices = ["A", "B"]
        G = Graph(vertices, ud_edges=ud_edges)
        self.assertTrue(Graph)
        # test empty BG
        Graph()

    def test_simple_edge_operations(self):
        vertices = ["A", "a", "B", "C", "D", "Y"]
        di_edges = [("A", "a"), ("B", "C")]
        bi_edges = [("D", "Y"), ("A", "Y")]
        ud_edges = [("B", "C"), ("C", "D")]
        G = Graph(
            vertices, di_edges=di_edges, bi_edges=bi_edges, ud_edges=ud_edges
        )

        # drawing tests (don't actually do anything)
        G.draw()
        G.draw(direction="LR")

        # test neighbour relations
        self.assertEqual(G.neighbors(["C"]), set({"B", "D"}))

        # test that you can delete di edges
        G.delete_diedge("A", "a")
        self.assertTrue(("A", "a") not in G.di_edges)
        self.assertTrue("a" not in G.children(["A"]))
        self.assertTrue("A" not in G.parents(["a"]))

        # test that you can delete bi edges
        self.assertTrue(G.has_biedge("Y", "A"))
        G.delete_biedge("A", "Y")
        self.assertTrue(("A", "Y") not in G.bi_edges)
        self.assertFalse(G.has_biedge("Y", "A"))
        self.assertTrue("A" not in G.siblings(["Y"]))
        self.assertTrue("Y" not in G.siblings(["A"]))

        # test that you can delete ud edges
        self.assertEqual(G.neighbors(["C"]), set({"B", "D"}))
        G.delete_udedge("D", "C")
        self.assertTrue("D" not in G.neighbors(["C"]))

    def test_add_same_edge_twice(self):
        vertices = ["A", "B"]
        di_edges = [("A", "B"), ("A", "B")]
        bi_edges = [("A", "B"), ("B", "A")]
        G = Graph(
            vertices, di_edges=di_edges, bi_edges=bi_edges, ud_edges=list()
        )
        assert len(G.bi_edges) == 1

    def test_directed_paths(self):

        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [
            ("A", "B"),
            ("A", "D"),
            ("B", "C"),
            ("C", "Y"),
            ("B", "D"),
            ("D", "Y"),
        ]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = Graph(vertices, di_edges, bi_edges)

        # test there are no directed paths from Y to A
        self.assertEqual([], G.directed_paths(["Y"], ["A"]))
        A_to_Y_paths = G.directed_paths(["A"], ["Y"])
        # TODO: How to assert equality of these paths
        true_paths = [
            [("A", "D"), ("D", "Y")],
            [("A", "B"), ("B", "D"), ("D", "Y")],
            [("A", "B"), ("B", "C"), ("C", "Y")],
        ]
        print(A_to_Y_paths)

    def test_subgraph(self):

        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("C", "D")]
        ud_edges = [("C", "D")]
        bi_edges = [("A", "Y"), ("A", "C"), ("A", "B")]
        G = Graph(vertices, di_edges, bi_edges, ud_edges)
        G.vertices["A"].fixed = True

        Gsub = G.subgraph(["A", "B", "Y"])
        self.assertEqual(set(["A", "B", "Y"]), set(Gsub.vertices))
        self.assertEqual(set([("A", "B")]), Gsub.di_edges)
        self.assertEqual(set([]), Gsub.ud_edges)
        self.assertEqual(set([("A", "Y"), ("A", "B")]), Gsub.bi_edges)

    def test_topological_sort(self):

        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("B", "C"), ("D", "C"), ("C", "Y")]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = Graph(vertices, di_edges, bi_edges)
        top_order = G.topological_sort()
        self.assertTrue(
            top_order == ["A", "D", "B", "C", "Y"]
            or top_order == ["D", "A", "B", "C", "Y"]
            or top_order == ["A", "B", "D", "C", "Y"]
        )

    def test_pre(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("B", "C"), ("D", "C"), ("C", "Y")]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = Graph(vertices, di_edges, bi_edges)
        top_order = ["D", "A", "B", "Y"]
        self.assertEqual(G.pre(["D", "A"], top_order), [])
        self.assertEqual(G.pre(["B", "C"], top_order), ["D", "A"])

    def test_delete_vertex(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("C", "D")]
        ud_edges = [("C", "D")]
        bi_edges = [("A", "Y"), ("A", "C"), ("A", "B")]
        G = Graph(vertices, di_edges, bi_edges, ud_edges)
        G.vertices["A"].fixed = True

        G.delete_vertex("B")
        assert "B" not in G.vertices
        assert ("A", "B") not in G.di_edges
        assert ("A", "B") not in G.bi_edges

    def test_subgraph(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("C", "D")]
        ud_edges = [("C", "D")]
        bi_edges = [("A", "Y"), ("A", "C"), ("A", "B")]
        G = Graph(vertices, di_edges, bi_edges, ud_edges)
        G.vertices["A"].fixed = True
        sub = G.subgraph(["A", "B", "C", "D"])
        assert set(sub.di_edges) == {("A", "B"), ("C", "D")}
        assert set(sub.ud_edges) == {("C", "D")}
        assert set(sub.bi_edges) == {("A", "C"), ("A", "B")}
        assert sub.vertices["A"].fixed == True


if __name__ == "__main__":
    unittest.main()



import unittest

from ananke.graphs import BG


class TestBG(unittest.TestCase):
    def test_bg_can_be_initialised(self):
        bi_edges = [("A", "B")]
        vertices = ["A", "B"]
        bg = BG(vertices, bi_edges=bi_edges)
        self.assertTrue(BG)
        # test empty BG
        BG()


if __name__ == "__main__":
    unittest.main()



import unittest

import ananke
from ananke.graphs import ADMG, admg


class TestADMG(unittest.TestCase):
    def test_obtaining_districts(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [
            ("A", "B"),
            ("A", "D"),
            ("B", "C"),
            ("C", "Y"),
            ("B", "D"),
            ("D", "Y"),
        ]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = ADMG(vertices, di_edges, bi_edges)
        self.assertCountEqual(([{"C", "A"}, {"Y", "B", "D"}]), (G.districts))
        self.assertEqual({"C", "A"}, G.district("A"))

    def test_obtaining_districts_large(self):
        vertices = ["X1", "U", "X2", "A1", "A2", "Y1", "Y2"]
        di_edges = [
            ("X1", "A1"),
            ("X1", "Y1"),
            ("A1", "Y1"),
            ("X2", "A2"),
            ("X2", "Y2"),
            ("A2", "Y2"),
            ("U", "A1"),
            ("U", "Y1"),
            ("U", "A2"),
            ("U", "Y2"),
            ("A2", "Y1"),
            ("A1", "Y2"),
        ]
        bi_edges = [("X1", "U"), ("U", "X2"), ("X1", "X2"), ("Y1", "Y2")]
        G = ADMG(vertices, di_edges, bi_edges)
        self.assertCountEqual(
            ([{"X1", "X2", "U"}, {"A1"}, {"A2"}, {"Y1", "Y2"}]), (G.districts)
        )
        self.assertEqual(G.district("X2"), {"X1", "X2", "U"})

    def test_fixed(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G = ADMG(vertices, di_edges, bi_edges)
        G.fix(["X_1"])
        self.assertEqual(G.fixed, ["X_1"])

    def test_is_subgraph(self):
        vertices = ["X_1", "W"]
        di_edges = [
            ("X_1", "W"),
        ]
        bi_edges = [("X_1", "W")]
        G = ADMG(vertices, di_edges, bi_edges)
        G.fix(["X_1"])
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G1 = ADMG(vertices, di_edges, bi_edges)
        self.assertTrue(G.is_subgraph(G1))

    def test_is_ancestral_subgraph(self):
        vertices = ["X_1", "W"]
        di_edges = [("X_1", "W")]
        bi_edges = [("X_1", "W")]
        G = ADMG(vertices, di_edges, bi_edges)
        G.fix(["X_1"])
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G1 = ADMG(vertices, di_edges, bi_edges)
        G1.fix(["X_1"])
        self.assertTrue(G.is_ancestral_subgraph(G1))

        vertices = ["X_2", "W", "Y"]
        di_edges = [("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_2", "Y")]
        G = ADMG(vertices, di_edges, bi_edges)
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G1 = ADMG(vertices, di_edges, bi_edges)
        G1.fix(["X_1"])
        self.assertFalse(G.is_ancestral_subgraph(G1))

    def test_subgraph(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("C", "D")]
        bi_edges = [("A", "Y"), ("A", "C"), ("A", "B")]
        G = ADMG(vertices, di_edges, bi_edges)
        G.vertices["A"].fixed = True

        Gsub = G.subgraph(["A", "B", "Y"])
        self.assertEqual(set(["A", "B", "Y"]), set(Gsub.vertices))
        self.assertEqual(set([("A", "B")]), Gsub.di_edges)
        self.assertEqual(set([("A", "Y"), ("A", "B")]), Gsub.bi_edges)

    def test_fixing(self):
        G = ADMG(
            vertices=["A", "B", "C"],
            di_edges=[("A", "B"), ("B", "C")],
            bi_edges=[("B", "C")],
        )
        G.fix(["C"])
        self.assertTrue(("B", "C") not in G.di_edges)
        self.assertTrue(("B", "C") not in G.bi_edges)
        self.assertTrue(G.vertices["C"].fixed)

    def test_reachable_closure(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B"), ("C", "B"), ("C", "A")]
        bi_edges = [("A", "B")]
        G = ADMG(vertices=vertices, di_edges=di_edges, bi_edges=bi_edges)
        cl, _, _ = G.reachable_closure(["B"])
        self.assertEqual({"A", "B"}, set(cl))

    def test_marg_dag_projection(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B"), ("B", "C")]
        bi_edges = [("B", "C")]
        G = ADMG(vertices=vertices, di_edges=di_edges, bi_edges=bi_edges)
        marg = G.maximal_arid_projection()
        self.assertEqual(
            set([("A", "B"), ("B", "C"), ("A", "C")]), marg.di_edges
        )
        self.assertEqual(set([]), marg.bi_edges)

    def test_marg_biedge_projection(self):
        vertices = ["A", "B", "C"]
        di_edges = [("B", "A"), ("B", "C")]
        bi_edges = [("A", "B"), ("B", "C")]
        G = ADMG(vertices=vertices, di_edges=di_edges, bi_edges=bi_edges)
        marg = G.maximal_arid_projection()
        self.assertEqual(set([("B", "A"), ("B", "C")]), marg.di_edges)
        self.assertEqual(set([("A", "C")]), marg.bi_edges)

    def test_marg_4var_graph(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("A", "C"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "C"), ("B", "D")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        marg = G.maximal_arid_projection()
        self.assertEqual(
            set([("A", "B"), ("B", "C"), ("C", "D"), ("A", "C")]), marg.di_edges
        )
        self.assertEqual(set([("B", "D")]), marg.bi_edges)

    def test_markov_pillow(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("B", "C"), ("D", "C"), ("C", "Y")]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = ADMG(vertices, di_edges, bi_edges)
        top_order = ["D", "A", "B", "C", "Y"]
        self.assertEqual(G.markov_pillow(["A", "D"], top_order), set())
        self.assertEqual(
            G.markov_pillow(["C"], top_order), set(["A", "B", "D"])
        )

    def test_markov_blanket(self):
        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [("A", "B"), ("B", "C"), ("D", "C"), ("C", "Y")]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = ADMG(vertices, di_edges, bi_edges)
        self.assertEqual(G.markov_blanket(["A", "D"]), set(["C", "B", "Y"]))
        self.assertEqual(G.markov_blanket(["C"]), set(["A", "B", "D"]))

    def test_nps(self):
        vertices = ["Treatment", "M", "L", "Confounders", "Outcome"]
        di_edges = [
            ("Confounders", "M"),
            ("Confounders", "L"),
            ("Treatment", "M"),
            ("Treatment", "Outcome"),
            ("Treatment", "L"),
            ("M", "L"),
            ("L", "Outcome"),
        ]
        bi_edges_G1 = [
            ("Treatment", "Confounders"),
            ("M", "Outcome"),
            ("L", "Outcome"),
        ]
        bi_edges_G2 = [("Treatment", "Confounders"), ("M", "Outcome")]

        G1 = ADMG(vertices, di_edges, bi_edges_G1)
        G2 = ADMG(vertices, di_edges, bi_edges_G2)

        self.assertTrue(G1.nonparametric_saturated())
        self.assertFalse(G2.nonparametric_saturated())

        vertices = ["A", "B", "C"]
        di_edges = [("B", "A"), ("B", "C")]
        bi_edges = [("A", "B"), ("C", "B")]
        G3 = ADMG(vertices, di_edges, bi_edges)

        self.assertTrue(G3.nonparametric_saturated())

    def test_mseparation_chain(self):
        vertices = ["A", "B", "C", "D", "E"]
        di_edges = [("E", "D")]
        bi_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        G = ADMG(vertices, di_edges, bi_edges)

        self.assertFalse(G._m_separated("A", "B", ["D"]))
        self.assertFalse(G._m_separated("A", "B", ["E"]))
        self.assertTrue(G._m_separated("A", "D", ["C"]))
        self.assertTrue(G._m_separated("A", "E", ["B", "C"]))

    def test_m_separation_fixed(self):
        vertices = ["A", "B", "C", "D", "E"]
        di_edges = [("A", "B"), ("C", "E")]
        bi_edges = [("B", "D"), ("D", "E")]
        G2 = ADMG(vertices, di_edges, bi_edges)
        G2.fix("A")
        G2.fix("C")

        self.assertFalse(G2.m_separated("A", "D", ["B"]))
        self.assertFalse(G2.m_separated("A", "B", ["E"]))

    def test_m_separation_no_bidirected(self):
        vertices = ["A", "B", "C", "D", "E"]
        di_edges = [("A", "B"), ("C", "D"), ("D", "E"), ("D", "B")]
        bi_edges = []
        G3 = ADMG(vertices, di_edges, bi_edges)

        self.assertTrue(G3._m_separated("C", "B", ["D"]))
        self.assertFalse(G3._m_separated("A", "B", ["C"]))
        self.assertFalse(G3._m_separated("A", "D", ["B"]))

    def test_m_separation_simple(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("D", "C")]
        bi_edges = [("B", "C")]
        G4 = ADMG(vertices, di_edges, bi_edges)

        self.assertTrue(G4._m_separated("A", "D", []))
        self.assertFalse(G4._m_separated("A", "D", ["B", "C"]))
        self.assertFalse(G4._m_separated("A", "C", ["B"]))

    def test_m_separation_sets(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "B")]
        G4 = ADMG(vertices, di_edges, bi_edges)
        self.assertTrue(G4.m_separated(["A", "B"], ["D"], ["C"]))
        self.assertFalse(G4.m_separated("A", ["B", "D"], ["C"]))

    def test_m_separation_undefined(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "B")]
        G4 = ADMG(vertices, di_edges, bi_edges)
        G4.fix(["A", "D"])
        with self.assertRaises(ananke.graphs.admg.UndefinedADMGOperation) as f:
            G4.m_separated("A", "D")

    def test_m_separation_fixed_vertices(self):
        vertices = ["A", "B", "C", "D", "E"]
        di_edges = [("A", "C"), ("C", "E"), ("D", "E"), ("B", "D")]
        bi_edges = []
        G4 = ADMG(vertices, di_edges, bi_edges)
        G4.fix(["A", "B"])

        self.assertTrue(G4.m_separated("C", "D"))

    def test_m_separation_fixed_vertices_complicated(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("B", "C"), ("C", "D")]
        bi_edges = [("A", "D"), ("A", "C")]
        G = ADMG(vertices, di_edges, bi_edges)
        G.fix("B")

        self.assertTrue(G.m_separated("B", "D", ["C"]))
        self.assertFalse(G.m_separated("B", "D", ["C", "A"]))

    def test_latent_projection(self):
        vertices = ["A", "B", "C"]
        di_edges = [("A", "B"), ("B", "C")]
        G = ADMG(vertices, di_edges, list())

        G1 = G.latent_projection(retained_vertices=["A", "C"])

        assert sorted(list(G1.vertices)) == sorted(["A", "C"])
        assert G1.di_edges == {("A", "C")}
        assert G1.bi_edges == set()

        vertices = ["A", "B", "C"]
        di_edges = [("B", "A"), ("B", "C")]
        G = ADMG(vertices, di_edges, list())

        G2 = G.latent_projection(["A", "C"])

        assert sorted(list(G2.vertices)) == sorted(["A", "C"])
        assert G2.di_edges == set()
        assert G2.bi_edges == {("A", "C")}

        vertices = ["A", "B", "C", "D", "E"]
        di_edges = [("B", "A"), ("B", "C"), ("D", "C"), ("D", "E")]
        G = ADMG(vertices, di_edges, list())

        G3 = G.latent_projection(["A", "C", "E"])

        assert G3.bi_edges == {("A", "C"), ("C", "E")}

        vertices = ["A", "B", "C", "D", "E", "F"]
        di_edges = [
            ("C", "A"),
            ("C", "D"),
            ("B", "A"),
            ("B", "D"),
            ("E", "F"),
            ("F", "D"),
        ]
        bi_edges = [("A", "E")]
        G = ADMG(vertices, di_edges, bi_edges)

        G4 = G.latent_projection(["A", "D", "F"])

        assert G4.bi_edges == {("A", "F"), ("A", "D")}

    def test_canonical_dag(self):
        vertices = ["A", "B", "C"]
        di_edges = [("B", "A"), ("B", "C")]
        bi_edges = [("B", "A"), ("B", "C")]
        G = ADMG(vertices, di_edges, bi_edges)

        G_canonical = G.canonical_dag()

        assert set(G_canonical.vertices) == {"A", "B", "C", "U_A_B", "U_B_C"}
        assert G_canonical.bi_edges == set()


if __name__ == "__main__":
    unittest.main()



"""
Class for one line ID algorithms.
"""

import copy
import functools
import itertools
import logging
import math
import os
from collections import ChainMap, namedtuple
from typing import Dict, List, NamedTuple, Union

import numpy as np
import pgmpy
import sympy as sp
from pgmpy.factors.discrete import DiscreteFactor, TabularCPD
from pgmpy.inference import VariableElimination

from ..factors import SymCPD, SymDiscreteFactor
from ..graphs import DAG, Vertex
from ..inference.variable_elimination import variable_elimination
from ..models import bayesian_network
from ..models.bayesian_network import BayesianNetwork, intervene

logger = logging.getLogger(__name__)


class NotIdentifiedError(Exception):
    """
    Custom error for when desired functional is not identified.
    """

    pass


class OneLineID:
    def __init__(self, graph, treatments, outcomes):
        """
        Applies the ID algorithm (Shpitser and Pearl, 2006) reformulated in a 'one-line' fashion (Richardson et al., 2017).

        :param graph: Graph on which the query will run.
        :param treatments: iterable of names of variables being intervened on.
        :param outcomes: iterable of names of variables whose outcomes we are interested in.
        """

        self.graph = graph
        self.treatments = [A for A in treatments]
        self.outcomes = [Y for Y in outcomes]
        self.swig = copy.deepcopy(graph)
        self.swig.fix(self.treatments)
        self.ystar = {
            v
            for v in self.swig.ancestors(self.outcomes)
            if not self.swig.vertices[v].fixed
        }
        self.Gystar = self.graph.subgraph(self.ystar)
        # dictionary mapping the fixing order for each p(D | do(V\D) )
        self.fixing_orders = {}

    def draw_swig(self, direction=None):
        """
        Draw the proper SWIG corresponding to the causal query.

        :return: dot language representation of the SWIG.
        """

        swig = copy.deepcopy(self.graph)

        # add fixed vertices for each intervention
        for A in self.treatments:

            fixed_vertex_name = A.lower()
            swig.add_vertex(fixed_vertex_name)
            swig.vertices[fixed_vertex_name].fixed = True

            # delete all outgoing edges from random vertex
            # and give it to the fixed vertex
            for edge in swig.di_edges:
                if edge[0] == A:
                    swig.delete_diedge(edge[0], edge[1])
                    swig.add_diedge(fixed_vertex_name, edge[1])

            # finally, add a fake edge between interventions and fixed vertices
            # just for nicer visualization
            swig.add_diedge(A, A.lower())

        return swig.draw(direction)

    def id(self):
        """
        Run one line ID for the query.

        :return: boolean that is True if p(Y(a)) is ID, else False.
        """

        self.fixing_orders = {}
        vertices = set(self.graph.vertices)

        # check if each p(D | do(V\D) ) corresponding to districts in Gystar is ID
        for district in self.Gystar.districts:

            fixable, order = self.graph.fixable(vertices - district)

            # if any piece is not ID, return not ID
            if not fixable:
                return False

            self.fixing_orders[tuple(district)] = order

        return True

    # TODO try to reduce functional
    def functional(self):
        """
        Creates and returns a string for identifying functional.

        :return: string representing the identifying functional.
        """

        if not self.id():
            raise NotIdentifiedError

        # create and return the functional
        functional = "" if set(self.ystar) == set(self.outcomes) else "\u03A3"

        for y in self.ystar:
            if y not in self.outcomes:
                functional += y
        if len(self.ystar) > 1:
            functional += " "

        print(self.fixing_orders)
        for district in sorted(list(self.fixing_orders)):
            functional += (
                "\u03A6"
                + "".join(reversed(self.fixing_orders[district]))
                + "(p(V);G) "
            )
        return functional

    # TODO export intermediate CADMGs for visualization
    def export_intermediates(self, folder="intermediates"):
        """
        Export intermediate CADMGs obtained during fixing.

        :param folder: string specifying path to folder where the files will be written.
        :return: None.
        """

        # make the folder if it doesn't exist
        if not os.path.exists(folder):
            os.mkdir(folder)

        # clear the directory
        for f in os.listdir(folder):
            file_path = os.path.join(folder, f)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception as e:
                print(e)

        # do the fixings and render intermediate CADMGs
        for district in self.fixing_orders:

            G = copy.deepcopy(self.graph)

            fixed_vars = ""
            dis_name = "".join(district)

            for v in self.fixing_orders[district]:
                fixed_vars += v
                G.fix(v)
                G.draw().render(
                    os.path.join(
                        folder, "phi" + fixed_vars + "_dis" + dis_name + ".pdf"
                    )
                )


def get_required_intrinsic_sets(admg):
    required_intrinsic_sets, _ = admg.get_intrinsic_sets()
    return required_intrinsic_sets


def get_allowed_intrinsic_sets(experiments):
    allowed_intrinsic_sets = set()
    allowed_intrinsic_dict = dict()
    fixing_orders = dict()
    for index, experiment in enumerate(experiments):
        intrinsic_sets, order_dict = experiment.get_intrinsic_sets()
        allowed_intrinsic_sets.update(intrinsic_sets)
        fixing_orders[index] = order_dict
        for s in intrinsic_sets:
            allowed_intrinsic_dict[frozenset(s)] = index
    return allowed_intrinsic_sets, allowed_intrinsic_dict, fixing_orders


def check_experiments_ancestral(admg, experiments):
    """
    Check that each experiment G(S(b_i)) is ancestral in ADMG G(V(b_i))
    https://simpleflying.com/

    :param admg: An ADMG
    :param experiments: A list of ADMGs representing experiments
    :return:
    """
    for experiment in experiments:
        graph = copy.deepcopy(admg)
        fixed = experiment.fixed
        graph.fix(fixed)
        if not experiment.is_ancestral_subgraph(admg):
            return False

    return True


def check_experiments_conform_to_gid(admg, experiments):
    graph = copy.deepcopy(admg)
    for experiment in experiments:
        if set(graph.vertices) != set(experiment.vertices):
            return False
    return True


class OneLineAID:
    def __init__(self, graph, treatments, outcomes):
        """
        Applies the one-line AID algorithm.

        :param graph: Graph on which the query will be run
        :param treatments: Iterable of treatment variables
        :param outcomes: Iterable of outcome variables
        """
        self.graph = graph
        self.treatments = treatments
        self.outcomes = outcomes
        self.swig = copy.deepcopy(graph)
        self.swig.fix(self.treatments)
        self.ystar = {
            v
            for v in self.swig.ancestors(self.outcomes)
            if not self.swig.vertices[v].fixed
        }
        self.Gystar = self.graph.subgraph(self.ystar)

        self.checked_id = False

    def id(self, experiments):
        """
        Checks if identification query is identified given the set of experimental distributions.

        :param experiments: a list of ADMG objects in which intervened variables are fixed.
        """
        if not check_experiments_ancestral(
            admg=self.graph, experiments=experiments
        ):

            raise NotIdentifiedError

        return self._id(experiments)

    def _id(self, experiments):
        self.required_intrinsic_sets = get_required_intrinsic_sets(
            admg=self.Gystar
        )
        (
            self.allowed_intrinsic_sets,
            self.allowed_intrinsic_dict,
            self.fixing_orders,
        ) = get_allowed_intrinsic_sets(experiments=experiments)

        is_id = False
        if self.allowed_intrinsic_sets >= self.required_intrinsic_sets:
            is_id = True

        self.checked_id = True

        return is_id

    def functional(self, experiments):
        """
        Creates a string representing the identifying functional

        :param experiments: A list of sets denoting the interventions of the available experimental distributions
        :return:
        """
        # if not check_experiments_ancestral(admg=self.graph, experiments=experiments):
        #    raise NotIdentifiedError
        if not self.id(experiments=experiments):
            raise NotIdentifiedError

        # create and return the functional
        functional = "" if set(self.ystar) == set(self.outcomes) else "\u03A3"

        for y in self.ystar:
            if y not in self.outcomes:
                functional += y
        if len(self.ystar) > 1:
            functional += " "

        # guarantee a deterministic printing order
        fixing = []
        intrinsic_sets = []
        for intrinsic_set in self.required_intrinsic_sets:
            fixed = experiments[
                self.allowed_intrinsic_dict[intrinsic_set]
            ].fixed
            fixing.append(list(fixed))
            intrinsic_sets.append(intrinsic_set)

        sorted_intrinsic_sets = sorted(
            intrinsic_sets, key=dict(zip(intrinsic_sets, fixing)).get
        )
        sorted_fixing = sorted(fixing)

        for i, intrinsic_set in enumerate(sorted_intrinsic_sets):
            fixed = sorted_fixing[i]
            vars = sorted(
                set(
                    [
                        v
                        for v in experiments[
                            self.allowed_intrinsic_dict[intrinsic_set]
                        ].vertices
                    ]
                )
                - set(fixed)
            )
            correct_order = self.fixing_orders[
                self.allowed_intrinsic_dict[intrinsic_set]
            ][frozenset(intrinsic_set) - frozenset(fixed)]
            if len(correct_order):
                functional += "\u03A6" + ",".join(reversed(correct_order))
            functional += " p({0} | do({1}))".format(
                ",".join(vars), ",".join(fixed)
            )

        return functional


class OneLineGID(OneLineAID):
    def __init__(self, graph, treatments, outcomes):
        """
        Applies the naive one-line GID algorithm.

        :param graph: Graph on which the query will be run.
        :param interventions: Iterable of treatment variables.
        :param outcomes: Iterable of outcome variables.
        """
        super().__init__(graph, treatments, outcomes)

    def id(self, experiments=list()):
        """
        Checks if identification query is identified given the set of experimental distributions.

        :param experiments: A list of ADMG objects denoting the interventions of the available experimental distributions.
        :return: boolean indicating if query is ID or not.
        """
        if not check_experiments_conform_to_gid(
            admg=self.graph, experiments=experiments
        ):

            raise NotIdentifiedError

        return self._id(experiments)


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


def _assert_valid_ananke_witness(
    net_1, net_2, observed_variables, treatment_dict, outcome_variables
):
    marginal_1 = variable_elimination(net_1, observed_variables)
    marginal_2 = variable_elimination(net_2, observed_variables)

    for v in observed_variables:
        assert net_1.vertices[v].cardinality == net_2.vertices[v].cardinality

    for val in itertools.product(
        *[range(net_1.vertices[v].cardinality) for v in observed_variables]
    ):
        val_dict = dict(zip(observed_variables, val))
        if isinstance(marginal_1, SymDiscreteFactor):
            val1 = marginal_1.get_value(**val_dict)
            val2 = marginal_2.get_value(**val_dict)
            logger.debug(f"{val_dict}, {val1}, {val2}")

            assert (
                sp.simplify(val1 - val2) == 0
            ), f"Observed distributions disagree for {val_dict}: {val1} != {val2}"
        elif isinstance(marginal_1, DiscreteFactor):
            val_dict = dict(zip(observed_variables, val))
            val1 = marginal_1.get_value(**val_dict)
            val2 = marginal_2.get_value(**val_dict)
            logger.debug(f"{val_dict}, {val1}, {val2}")

            assert math.isclose(
                val1, val2
            ), f"Observed distributions disagree for {val_dict}: {val1} != {val2}"

    intervened_marginal_1 = variable_elimination(
        copy.deepcopy(net_1).fix(treatment_dict), outcome_variables
    )
    intervened_marginal_2 = variable_elimination(
        copy.deepcopy(net_2).fix(treatment_dict), outcome_variables
    )
    diff = list()
    for val in itertools.product(
        *[range(net_1.vertices[v].cardinality) for v in outcome_variables]
    ):
        val_dict = dict(zip(outcome_variables, val))
        diff.append(
            intervened_marginal_1.get_value(**val_dict)
            - intervened_marginal_2.get_value(**val_dict)
        )

    assert diff != [0] * len(
        diff
    ), "Counterfactual distributions agree between models"


def _assert_valid_pgmpy_witness(
    net_1, net_2, observed_variables, treatment_dict, outcome_variables
):
    inference_1 = VariableElimination(net_1)
    inference_2 = VariableElimination(net_2)
    marginal_1 = inference_1.query(observed_variables)
    marginal_2 = inference_2.query(observed_variables)

    for v in observed_variables:
        assert net_1.get_cardinality(v) == net_2.get_cardinality(v)

    for val in itertools.product(
        *[range(net_1.get_cardinality(v)) for v in observed_variables]
    ):
        val_dict = dict(zip(observed_variables, val))
        val1 = marginal_1.get_value(**val_dict)
        val2 = marginal_2.get_value(**val_dict)
        logger.debug(f"{val_dict}, {val1}, {val2}")

        assert math.isclose(
            val1, val2
        ), f"Observed distributions disagree for {val_dict}: {val1} != {val2}"

    intervened_net_1 = intervene(net_1, treatment_dict)
    intervened_net_2 = intervene(net_2, treatment_dict)
    intervened_inference_1 = VariableElimination(intervened_net_1)
    intervened_inference_2 = VariableElimination(intervened_net_2)
    intervened_marginal_1 = intervened_inference_1.query(outcome_variables)
    intervened_marginal_2 = intervened_inference_2.query(outcome_variables)

    m1 = list()
    m2 = list()
    for val in itertools.product(
        *[range(net_1.get_cardinality(v)) for v in outcome_variables]
    ):
        val_dict = dict(zip(outcome_variables, val))
        m1.append(intervened_marginal_1.get_value(**val_dict))
        m2.append(intervened_marginal_2.get_value(**val_dict))

    assert m1 != m2, "Counterfactual distributions agree between models"


def assert_valid_witness(
    net_1: Union[
        pgmpy.models.BayesianNetwork, bayesian_network.BayesianNetwork
    ],
    net_2: Union[
        pgmpy.models.BayesianNetwork, bayesian_network.BayesianNetwork
    ],
    observed_variables: list,
    treatment_dict: dict,
    outcome_variables=None,
):
    """
    Asserts that two BayesianNetwork objects represent a valid witness for identification, meaning
    that the two Bayesian networks agree on the marginal distribution over `observed_variables` but disagree in
    at least one part of the counterfactual distribution for `outcome_variables` under the
    intervention specified by `treatment_dict`.

    :param net_1: The first BayesianNetwork object
    :param net_2: The second BayesianNetwork object
    :param observed_variables: A list of variables for the observed margin
    :param treatment_dict: A dictionary of treatment variables and values
    :param outcome_variables: An optional list of outcome variables. If left unspecified, then it is
    all variables in `observed_variables` except those in `treatment_dict`.
    """

    if outcome_variables is None:
        outcome_variables = [
            v for v in observed_variables if v not in treatment_dict
        ]
    if isinstance(net_1, pgmpy.models.BayesianNetwork) and isinstance(
        net_2, pgmpy.models.BayesianNetwork
    ):
        _assert_valid_pgmpy_witness(
            net_1, net_2, observed_variables, treatment_dict, outcome_variables
        )
    elif isinstance(net_1, bayesian_network.BayesianNetwork) and isinstance(
        net_2, bayesian_network.BayesianNetwork
    ):
        _assert_valid_ananke_witness(
            net_1, net_2, observed_variables, treatment_dict, outcome_variables
        )
    else:
        raise ValueError("Mismatched or unrecognized Bayesian Network objects")



import unittest

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_allclose

from ananke.graphs import ADMG
from ananke.models import LinearGaussianSEM


class TestLinearGaussianSEM(unittest.TestCase):
    def test_model_creation(self):

        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        bi_edges = [("B", "D")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        model = LinearGaussianSEM(G)
        self.assertEqual(8, model.n_params)

    def test_verma_model(self):

        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        bi_edges = [("B", "D")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        model = LinearGaussianSEM(G)

        # generate data from an ADMG A->B->C->D B<->D and try to fit
        N = 50000
        dim = 4

        omega = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0.8], [0, 0, 1, 0], [0, 0.8, 0, 1]]
        )

        beta = np.array(
            [[0, 0, 0, 0], [3, 0, 0, 0], [0, -1, 0, 0], [0, 0, 2.5, 0]]
        )

        # generate data
        true_sigma = (
            np.linalg.inv(np.eye(dim) - beta)
            @ omega
            @ np.linalg.inv((np.eye(dim) - beta).T)
        )
        X = np.random.multivariate_normal([0] * dim, true_sigma, size=N)
        data = pd.DataFrame(
            {"A": X[:, 0], "B": X[:, 1], "C": X[:, 2], "D": X[:, 3]}
        )

        # test that without fitting you can't compute likelihood
        with self.assertRaises(AssertionError):
            model.neg_loglikelihood(data)

        # test that without fitting you can't draw the model
        with self.assertRaises(AssertionError):
            model.draw()

        # try with trust-exact (the default)
        model = LinearGaussianSEM(G)
        model.fit(data)

        # computation of causal effects
        self.assertEqual(0, model.total_effect(["D"], ["A"]))
        self.assertAlmostEqual(
            -7.5, model.total_effect(["A"], ["D"]), delta=0.5
        )

        model.draw(direction="LR")

    def test_that_beta_and_omega_recovered_correctly(self):
        np.random.seed(123)
        size = 10000
        dim = 4
        # generate data from A->B->C->D, A<->D, A<->C
        beta = np.array(
            [[0, 1.2, 0, 0], [0, 0, -1.5, 0], [0, 0, 0, 1.0], [0, 0, 0, 0]]
        ).T

        omega = np.array(
            [
                [1.2, 0, 0.5, 0.6],
                [0, 1, 0, 0.0],
                [0.5, 0, 1, 0],
                [0.6, 0.0, 0, 1],
            ]
        )

        true_sigma = (
            np.linalg.inv(np.eye(dim) - beta)
            @ omega
            @ np.linalg.inv((np.eye(dim) - beta).T)
        )
        X = np.random.multivariate_normal([0] * dim, true_sigma, size=size)
        X = X - np.mean(X, axis=0)  # centre the data
        data = pd.DataFrame(
            {"A": X[:, 0], "B": X[:, 1], "C": X[:, 2], "D": X[:, 3]}
        )

        # make Ananke ADMG and call RICF
        G = ADMG(
            vertices=["A", "B", "C", "D"],
            di_edges=[("A", "B"), ("B", "C"), ("C", "D")],
            bi_edges=[("A", "D"), ("A", "C")],
        )

        model = LinearGaussianSEM(G)
        model.fit(data, standardize=False)

        assert beta == pytest.approx(model.B_, rel=0.05)
        assert omega == pytest.approx(model.omega_, rel=0.05)

    def test_nested_markov_equivalent_models_are_bic_indistinguishable(self):
        # define number of samples and number of vertices
        np.random.seed(123)
        N = 1000
        dim = 4

        # define the Omega matrix
        omega = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0.8], [0, 0, 1, 0], [0, 0.8, 0, 1]]
        )

        # define the B matrix
        beta = np.array(
            [[0, 0, 0, 0], [3, 0, 0, 0], [1.2, -1, 0, 0], [0, 0, 2.5, 0]]
        )

        # generate data according to the graph
        true_sigma = (
            np.linalg.inv(np.eye(dim) - beta)
            @ omega
            @ np.linalg.inv((np.eye(dim) - beta).T)
        )
        X = np.random.multivariate_normal([0] * dim, true_sigma, size=N)
        data = pd.DataFrame(
            {"A": X[:, 0], "B": X[:, 1], "C": X[:, 2], "D": X[:, 3]}
        )

        # Set up the true model (arid projection of G)
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("A", "C"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "C"), ("B", "D")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        G_arid = G.maximal_arid_projection()

        # Set up incorrect model

        di_edges = [("A", "B"), ("B", "C"), ("C", "D"), ("B", "D"), ("A", "C")]
        bi_edges = []
        G_incorrect = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)

        # calculate BIC for incorrect model
        model_incorrect = LinearGaussianSEM(G_incorrect)
        model_incorrect.fit(data)
        bic_incorrect = model_incorrect.bic(data)

        # calculate BIC for the true model
        model_truth = LinearGaussianSEM(G_arid)
        model_truth.fit(data)
        bic_truth = model_truth.bic(data)

        assert bic_truth < bic_incorrect
        # calculate BIC for the nested Markov equivalent model (A -> B is now A <-> B)
        di_edges = [("A", "C"), ("B", "C"), ("C", "D")]
        bi_edges = [("A", "B"), ("B", "D")]
        G_equivalent = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        model_equivalent = LinearGaussianSEM(G_equivalent)
        model_equivalent.fit(data)
        bic_equivalent = model_equivalent.bic(data)

        assert model_equivalent.n_params == model_truth.n_params

        # assert BIC is the same for nested Markov equivalent model
        # note that since models have same dimension we can consider only the log-likelihood when checking equality of BIC
        assert model_truth.neg_loglikelihood(data) == pytest.approx(
            model_equivalent.neg_loglikelihood(data)
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import ananke
from ananke.graphs import ADMG, DAG
from ananke.identification import OptAdjustment


class TestComputeCounterfactual(unittest.TestCase):
    def test_generate_opt_adjustmen_case1(self):
        vertices = ["Q", "Z", "O", "E", "A", "M", "Y"]
        di_edges = [
            ("Q", "O"),
            ("Q", "A"),
            ("Z", "A"),
            ("Z", "E"),
            ("O", "M"),
            ("O", "Y"),
            ("E", "M"),
            ("A", "M"),
            ("M", "Y"),
        ]
        G = DAG(vertices, di_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        truth = {"E", "O"}
        assert truth == result

    def test_generate_opt_adjustmen_case2(self):
        vertices = ["Q", "Z", "O", "E", "A", "M", "Y", "L"]
        di_edges = [
            ("Q", "O"),
            ("Q", "A"),
            ("Z", "A"),
            ("Z", "E"),
            ("O", "M"),
            ("O", "Y"),
            ("E", "M"),
            ("A", "M"),
            ("M", "Y"),
            ("L", "M"),
            ("L", "Y"),
        ]
        G = DAG(vertices, di_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        truth = {"E", "O"}
        assert truth == result

    def test_generate_opt_adjustmen_case3(self):
        vertices = ["A", "C", "Y"]
        di_edges = [("A", "C"), ("Y", "C")]
        G = DAG(vertices, di_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        truth = []
        assert truth == result

    def test_generate_opt_adjustmen_case4(self):
        vertices = ["A", "Y"]
        di_edges = [("A", "Y")]
        G = DAG(vertices, di_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        assert result is None

    def test_generate_opt_adjustmen_case5(self):
        vertices = ["A", "Y", "C"]
        di_edges = [("A", "Y"), ("A", "C"), ("C", "Y")]
        G = DAG(vertices, di_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        assert result is None

    def test_admg_caes6(self):
        vertices = ["A", "M", "Y"]
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        G = ADMG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        model = OptAdjustment(G)
        result = model.generate_opt_adjustment_set("A", "Y")
        assert result is None



"""
Class to represent vertex of a graphical model.

The vertex class should be general enough to fit
into any kind of graphical model -- UGs, DAGs,
CGs, ADMGs, CPDAGs, CADMGs etc.
"""


class Vertex:
    def __init__(self, name, fixed=False, cardinality=2, attributes=None):
        """
        Constructor.

        :param name: name of the vertex.
        :param fixed: boolean specifying whether vertex is fixed or random.
        :param cardinality: integer indicating categories or string "continuous".
        """

        self.name = name
        self.parents = set()
        self.children = set()
        self.siblings = set()
        self.neighbors = set()
        self.fixed = fixed
        self.cardinality = cardinality
        self.attributes = attributes



from itertools import chain, combinations


def powerset(iterable, min_size=0):
    "powerset([1,2,3], 0) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(
        (combinations(s, r)) for r in range(min_size, len(s) + 1)
    )



"""
Class for Lauritzen-Wermuth-Frydenberg chain graphs (LWF-CGs/CGs).
"""

import logging

from .sg import SG

logger = logging.getLogger(__name__)


class CG(SG):
    def __init__(self, vertices=[], di_edges=set(), ud_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        :param ud_edges: iterable of tuples of undirected edges i.e. (X, Y) = X - Y.
        """

        # initialize vertices
        super().__init__(
            vertices=vertices, di_edges=di_edges, ud_edges=ud_edges, **kwargs
        )
        logger.debug("CG")

    def boundary(self, vertices):
        """
        Get the boundary of a set of vertices defined as the block and parents of the block.

        :param vertices: iterable of vertex names.
        :return: set corresponding to the boundary.
        """

        boundary = set()
        for v in vertices:
            boundary = boundary.union(self.block(v))
        boundary = boundary.union(self.parents(boundary))
        return boundary - set(vertices)



import unittest
from unittest.mock import patch

from ananke.graphs import SG


class TestSG(unittest.TestCase):
    def test_non_segregated_graph_raises_error(self):
        vertices = ["A", "B", "C"]
        bi_edges = [("A", "B")]
        ud_edges = [("B", "C")]
        with self.assertRaises(TypeError):
            G = SG(vertices, bi_edges=bi_edges, ud_edges=ud_edges)

    def test_directed_cycle_graph_raises_error(self):
        # directed cycle graph
        vertices = ["A", "B", "C"]
        bi_edges = [("A", "B")]
        di_edges = [("A", "B"), ("B", "C"), ("C", "A")]
        with self.assertRaises(TypeError):
            G = SG(vertices, bi_edges=bi_edges, di_edges=di_edges)

    def test_partially_directed_graph_raises_error(self):
        # partially directed cycle graph
        vertices = ["A", "B", "C"]
        bi_edges = []
        ud_edges = [("A", "B"), ("B", "C")]
        di_edges = [("C", "A")]
        with self.assertRaises(TypeError):
            G = SG(
                vertices,
                bi_edges=bi_edges,
                di_edges=di_edges,
                ud_edges=ud_edges,
            )

    def test_sg_does_not_raise_error(self):
        """
        This properly formed SG should not raise an AssertionError

        :return:
        """

        try:
            vertices = ["A", "B", "C"]
            bi_edges = []
            ud_edges = [("A", "B"), ("B", "C"), ("C", "A")]
            G = SG(vertices, bi_edges=bi_edges, ud_edges=ud_edges)
        except TypeError:
            self.fail("Correctly formed SG raised AssertionError")

    def test_biedge(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "Y"), ("X_2", "W")]
        ud_edges = [("X_1", "X_2")]
        G = SG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)
        G.add_biedge("X_1", "X_2")
        self.assertTrue("X_2" in G.siblings(["X_1"]))
        self.assertTrue("X_1" in G.siblings(["X_2"]))
        G.delete_biedge("X_1", "X_2")
        self.assertFalse("X_2" in G.siblings(["X_1"]))
        self.assertFalse("X_1" in G.siblings(["X_2"]))

    def test_udedge(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "Y"), ("X_2", "W")]
        ud_edges = [("X_1", "X_2")]
        G = SG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)
        G.add_udedge("W", "Y")
        self.assertTrue("W" in G.neighbors(["Y"]))
        self.assertTrue("Y" in G.neighbors(["W"]))
        G.delete_udedge("W", "Y")
        self.assertFalse("W" in G.neighbors(["Y"]))
        self.assertFalse("Y" in G.neighbors(["W"]))

    def test_that_districts_correct(self):

        vertices = ["A", "B", "C", "D", "Y"]
        di_edges = [
            ("A", "B"),
            ("A", "D"),
            ("B", "C"),
            ("C", "Y"),
            ("B", "D"),
            ("D", "Y"),
        ]
        bi_edges = [("A", "C"), ("B", "Y"), ("B", "D")]
        G = SG(vertices, di_edges, bi_edges)
        self.assertCountEqual(([{"A", "C"}, {"D", "Y", "B"}]), G.districts)
        self.assertEqual({"A", "C"}, G.district("A"))

    def test_genealogical_sets_correct(self):
        vertices = ["X1", "U", "X2", "A1", "A2", "Y1", "Y2"]
        di_edges = [
            ("X1", "A1"),
            ("X1", "Y1"),
            ("A1", "Y1"),
            ("X2", "A2"),
            ("X2", "Y2"),
            ("A2", "Y2"),
            ("U", "A1"),
            ("U", "Y1"),
            ("U", "A2"),
            ("U", "Y2"),
            ("A2", "Y1"),
            ("A1", "Y2"),
        ]
        bi_edges = [("X1", "U"), ("U", "X2"), ("X1", "X2"), ("Y1", "Y2")]
        G = SG(vertices, di_edges, bi_edges)
        self.assertCountEqual(
            (G.districts), [{"X2", "X1", "U"}, {"A1"}, {"A2"}, {"Y2", "Y1"}]
        )
        self.assertEqual(G.district("X2"), {"X2", "X1", "U"})
        self.assertEqual(G.ancestors(["A2"]), {"X2", "A2", "U"})
        self.assertEqual(G.ancestors(["A1"]), {"X1", "A1", "U"})
        self.assertEqual(
            G.ancestors(["A1", "A2"]), {"X2", "A1", "X1", "U", "A2"}
        )
        self.assertEqual(G.descendants(["A1", "A2"]), {"Y2", "Y1", "A1", "A2"})

    def test_fix_gives_correct_graph(self):

        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G = SG(vertices, di_edges, bi_edges)
        G.fix(["X_1"])
        self.assertTrue(G.vertices["X_1"].fixed)
        G_dis = {frozenset(s) for s in G.districts}

        self.assertEqual({frozenset({"X_2", "Y"}), frozenset({"W"})}, G_dis)
        self.assertEqual(
            sorted([("X_1", "W"), ("W", "Y"), ("X_2", "Y")]),
            sorted(list(G.di_edges)),
        )

        G.fix(["Y"])
        self.assertTrue(G.vertices["Y"].fixed)
        self.assertTrue("W" not in G.parents(["Y"]))
        self.assertTrue("X_2" not in G.siblings(["Y"]))

    @patch("ananke.graphs.sg.SG._calculate_districts")
    def test_districts_is_a_property(self, mock1):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        bi_edges = [("X_1", "W"), ("X_2", "Y"), ("X_1", "X_2")]
        G = SG(vertices, di_edges=di_edges, bi_edges=bi_edges)
        self.assertEqual(0, mock1.call_count)
        G.districts
        self.assertEqual(1, mock1.call_count)

    @patch("ananke.graphs.sg.SG._calculate_blocks")
    def test_blocks_is_a_property(self, mock1):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "W"), ("W", "Y"), ("X_2", "Y")]
        ud_edges = [("X_1", "W")]
        G = SG(vertices, di_edges=di_edges, ud_edges=ud_edges)

        self.assertEqual(0, mock1.call_count)
        G.blocks
        self.assertEqual(1, mock1.call_count)

    def test_block(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "Y"), ("X_2", "W")]
        ud_edges = [("X_1", "X_2")]
        G = SG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)
        self.assertEqual({"X_1", "X_2"}, G.block("X_1"))

    def test_blocks(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "Y"), ("X_2", "W")]
        ud_edges = [("X_1", "X_2")]
        G = SG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)
        self.assertEqual(
            {frozenset({"X_1", "X_2"}), frozenset({"W"}), frozenset({"Y"})},
            {frozenset(i) for i in G.blocks},
        )

    def test_fixable(self):
        vertices = ["X_1", "X_2", "W", "Y"]
        di_edges = [("X_1", "Y"), ("X_2", "W")]
        ud_edges = [("X_1", "X_2")]
        G = SG(vertices=vertices, di_edges=di_edges, ud_edges=ud_edges)
        fixable, order = G.fixable(["W", "X_1"])
        print(fixable, order)
        self.assertTrue(fixable)
        self.assertEqual(["W", "X_1"], order)


if __name__ == "__main__":
    unittest.main()



"""
Optimal adjustment sets finding
"""

import logging
import unittest

import numpy as np

import ananke
from ananke.graphs import ADMG, DAG

logger = logging.getLogger("opt_adjust")


def get_opt_set(G, vertex1, vertex2):

    """
    function to get the optimal set from two vertices

    :param G: the graph
    :param vertex1: inference from vertex1, say treatment
    :param vertex2: inference to vertex2, say outcome

    """

    causal_nodes = [
        item
        for item in G.descendants([vertex1])
        if item in G.ancestors([vertex2])
    ]
    if len(causal_nodes) == 0:
        logger.info("No causal node!")
        return None
    causal_nodes.remove(vertex1)
    forb_set = [item for item in G.descendants(causal_nodes)]
    forb_set = forb_set + [
        vertex1
    ]  # The forbidden set characterizes those covariates that may never be included into a valid adjustment set

    opt_set = [item for item in G.parents(causal_nodes) if item not in forb_set]
    return set(opt_set)


def get_opt_set_from_set(G, node_set1, node_set2):

    """
    function to get the optimal set from two vertices sets

    :param G: the graph
    :param node_set1: set of inference from
    :param node_set2: set of inference to

    """

    causal_nodes = [
        item
        for item in G.descendants(node_set1)
        if item in G.ancestors(node_set2)
    ]
    causal_nodes.remove(node_set1)
    forb_set = [item for item in G.descendants(causal_nodes)]
    forb_set = forb_set + [node_set1]
    input_set = [item for item in G.vertices if item not in forb_set]
    for element in input_set:
        new_set = [item for item in input_set if item not in element]
        condition_set = new_set + [vertex1]
        flag_new_set = G.d_separated(
            element, node_set2, condition_set
        )  # check if d-separated
        if flag_new_set:
            input_set = new_set
    return set(input_set)


def get_min_set(G, input, vertex1, vertex2):

    """
    function for minimal set

    :param G: DAG
    :param input: optimal set found using get_opt_set function
    :param vertex1: inference from vertex1, say treatment
    :param vertex2: inference to vertex2, say outcome

    """

    if len(input) == 0:
        logger.info("Adjustment set is empty!")
        return

    input_set = [item for item in input]

    is_minimal_set = 0  # flag for the minimal set

    for element in G.children(
        [vertex1]
    ):  # prepare for d-separation test, construct a SWIG
        G.delete_diedge(vertex1, element)

    for element in input_set:
        new_set = [item for item in input_set if item not in element]
        is_d_separated = G.d_separated(
            vertex1, vertex2, new_set
        )  # check if d-separated
        if is_d_separated:
            is_minimal_set = 1
            final_set = get_min_set(G, new_set, vertex1, vertex2)

    if not is_minimal_set:
        # input set is minimal set
        return set(input_set)

    return set(final_set)


class OptAdjustment:
    def __init__(self, graph):
        self.graph = graph

    def generate_opt_adjustment_set(self, vertex1, vertex2):

        """
        Fits the binary nested model. Let N the number of observations, M the number of variables.

        :param self: DAG
        :param vertex1: inference from vertex1, say treatment
        :param vertex2: inference to vertex2, say outcome

        """

        model = self.graph

        if type(model) == ADMG:
            logger.info("Found ADMG graph, not applicable")
            return None

        is_d_separated = model.d_separated(vertex1, vertex2, [])

        if is_d_separated:
            logger.info(
                "Tested vertices are independent so no set is needed to condition on"
            )
            return []

        opt_set = get_opt_set(model, vertex1, vertex2)

        if opt_set == None:
            logger.info("No adjustment set!")
            return None

        result = get_min_set(model, opt_set, vertex1, vertex2)

        return result



import numpy as np
import sympy as sp
from pgmpy.factors.discrete import TabularCPD

from ananke.factors import SymCPD
from ananke.graphs import DAG
from ananke.models import BayesianNetwork, bayesian_network


def test_fixing_in_bayesian_network_chain():
    x1 = sp.Symbol("x1")
    y1 = sp.Symbol("y1")
    y2 = sp.Symbol("y2")
    x2 = sp.Symbol("x2")
    x3 = sp.Symbol("x3")
    cpd_a = SymCPD("A", 2, sp.Array([[x1, y1], [1 - x1, 1 - y1]]), ["B"], [2])
    cpd_b = SymCPD("B", 2, sp.Array([[x2, y2], [1 - x2, 1 - y2]]), ["C"], [2])
    cpd_c = SymCPD("C", 2, sp.Array([[x3], [1 - x3]]))

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds = {"A": cpd_a, "B": cpd_b, "C": cpd_c}
    bayesian_network = BayesianNetwork(graph, cpds)
    bayesian_network.fix({"A": 0})

    assert bayesian_network.cpds["A"] == SymCPD(
        variable="A", variable_card=2, values=sp.Array([[1], [0]])
    )

    assert set(bayesian_network.di_edges) == set([("C", "B")])


def test_fixing_in_bayesian_network_collider():
    x1 = sp.Symbol("x1")
    y1 = sp.Symbol("y1")
    y2 = sp.Symbol("y2")
    y3 = sp.Symbol("y3")
    y4 = sp.Symbol("y4")
    x3 = sp.Symbol("x3")
    cpd_a = SymCPD("A", 2, sp.Array([[x1], [1 - x1]]))
    cpd_b = SymCPD(
        "B",
        2,
        sp.Array([[y1, y2, y3, y4], [1 - y1, 1 - y2, 1 - y3, 1 - y4]]),
        ["A", "C"],
        [2, 2],
    )
    cpd_c = SymCPD("C", 2, sp.Array([[x3], [1 - x3]]))

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds = {"A": cpd_a, "B": cpd_b, "C": cpd_c}
    net = BayesianNetwork(graph, cpds)
    net.fix({"B": 0})
    truth = SymCPD(
        variable="B",
        variable_card=2,
        values=sp.Array([[1], [0]]),
    )

    assert net.cpds["B"] == truth


def test_fixing_using_pgmpy_cpds():
    x1 = 0.3
    y1 = 0.2
    y2 = 0.3
    y3 = 0.5
    y4 = 0.7
    x3 = 0.2

    cpd_a = TabularCPD("A", 2, np.array([[x1], [1 - x1]]))
    cpd_b = TabularCPD(
        "B",
        2,
        np.array([[y1, y2, y3, y4], [1 - y1, 1 - y2, 1 - y3, 1 - y4]]),
        ["A", "C"],
        [2, 2],
    )
    cpd_c = TabularCPD("C", 2, np.array([[x3], [1 - x3]]))

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds = {"A": cpd_a, "B": cpd_b, "C": cpd_c}
    bayesian_network = BayesianNetwork(graph, cpds)
    bayesian_network.fix({"B": 0})
    truth = TabularCPD(
        variable="B",
        variable_card=2,
        values=np.array([[1], [0]]),
    )
    assert bayesian_network.cpds["B"] == truth


def test_bayesian_network_to_pgmpy():
    x1 = 0.3
    y1 = 0.2
    y2 = 0.3
    y3 = 0.5
    y4 = 0.7
    x3 = 0.2

    cpd_a = TabularCPD("A", 2, np.array([[x1], [1 - x1]]))
    cpd_b = TabularCPD(
        "B",
        2,
        np.array([[y1, y2, y3, y4], [1 - y1, 1 - y2, 1 - y3, 1 - y4]]),
        ["A", "C"],
        [2, 2],
    )
    cpd_c = TabularCPD("C", 2, np.array([[x3], [1 - x3]]))

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds = {"A": cpd_a, "B": cpd_b, "C": cpd_c}
    net = BayesianNetwork(graph, cpds)
    pgmpy_net = net.to_pgmpy()
    assert set(pgmpy_net.edges()) == set(graph.di_edges)
    assert set(pgmpy_net.nodes()) == set(graph.vertices)
    assert set(cpds.values()) == set(pgmpy_net.get_cpds())


def test_symbolic_cpd_generation():

    graph = DAG(
        vertices={"A": 2, "B": 2, "C": 2},
        di_edges=[("C", "B"), ("B", "A")],
    )
    cpds, all_vars = bayesian_network.create_symbolic_cpds(graph)
    assert set(all_vars) == {
        sp.Symbol("q_C_0"),
        sp.Symbol("q_B_0_0"),
        sp.Symbol("q_B_0_1"),
        sp.Symbol("q_A_0_0"),
        sp.Symbol("q_A_0_1"),
    }
    assert cpds["A"] == SymCPD(
        "A",
        2,
        sp.Array(
            [
                [sp.Symbol("q_A_0_0"), sp.Symbol("q_A_0_1")],
                [1 - sp.Symbol("q_A_0_0"), 1 - sp.Symbol("q_A_0_1")],
            ]
        ),
        ["B"],
        [2],
    )



"""
Helper functions that help load datasets in Ananke.
"""

import os

import pandas as pd

MODULE_PATH = os.path.dirname(__file__)


def load_conditionally_ignorable_data():
    """
    Load toy data for the conditionally ignorable model
    where the confounder is Viral Load, T is the treatment
    and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/conditionally_ignorable.csv")
    return pd.read_csv(path)


def load_afixable_data():
    """
    Load toy data for an adjustment fixable setting
    where T is the treatment and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/a_fixable.csv")
    return pd.read_csv(path)


def load_frontdoor_data():
    """
    Load toy data for frontdoor setting
    where T is the treatment and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/frontdoor.csv")
    return pd.read_csv(path)


def load_wisconsin_health_study():
    """
    Load the dataset extract from the Wisconsin Health Study presented in [1].
    Columns are defined as follows:
    X: an indicator of whether family income in 1957 was above \$5k;
    Y: an indicator of whether the respondents income in 1992 was above \$37k;
    M: an indicator of whether the respondent was drafted into the military;
    E: an indicator of whether the respondent had education beyond high school.
    count: the count of each event in (X, Y, M, E)

    [1] R. J. Evans and T. S. Richardson, “Smooth, identifiable supermodels of discrete DAG models with latent variables,” Bernoulli, vol. 25, no. 2, pp. 848–876, May 2019, doi: 10.3150/17-BEJ1005.


    """
    path = os.path.join(MODULE_PATH, "real/evans_richardson_wisconsin.csv")
    return pd.read_csv(path)
